# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edu>

from __future__ import unicode_literals

import logging

from six import text_type, itervalues, string_types
from collections import defaultdict

from ..command import Command
from ..expression import boolean

try:
    import xlsxwriter
except ImportError:
    xlsxwriter = None

logger = logging.getLogger(__name__)


class ExcelExportCommand(Command):
    """Export the metabolic model as an Excel workbook."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            'file', type=text_type,
            help='File path for writing the Excel workbook')

    def run(self):
        model = self._model
        if xlsxwriter is None:
            self.fail(
                'Excel export requires the XlsxWriter python module'
                ' ("pip install xlsxwriter")')

        workbook = xlsxwriter.Workbook(self._args.file)

        model_info = workbook.add_worksheet(name='Model Info')

        model_info.write(0, 0, ('Model Name: {}'.format(model.name)))
        model_info.write(
            1, 0, ('Biomass Reaction: {}'.format(model.biomass_reaction)))
        model_info.write(
            2, 0, ('Default Flux Limits: {}'.format(model.default_flux_limit)))
        if model.version_string is not None:
            model_info.write(
                3, 0, ('Version: {}'.format(model.version_string)))

        reaction_sheet = workbook.add_worksheet(name='Reactions')

        # get compound name dict
        compounds_name = {}
        for cpd in model.compounds:
            if cpd.name:
                compounds_name[cpd.id] = cpd.name
            else:
                compounds_name[cpd.id] = cpd.id

        property_set = set()
        for reaction in model.reactions:
            property_set.update(reaction.properties)
        property_list = list(property_set)
        property_list_sorted = sorted(property_list,
                                      key=lambda x: (x != 'id',
                                                     x != 'equation', x))
        model_reactions = set(model.model)
        for z, i in enumerate(property_list_sorted + ['in_model'] +
                              ['translated_equation']):
            reaction_sheet.write_string(0, z, text_type(i))
        gene_rxn = defaultdict(list)
        for x, i in enumerate(model.reactions):
            # get translated equation
            rx = i.equation
            translated_equation = str(rx.translated_compounds(
                lambda x: compounds_name.get(x, x)))

            for y, j in enumerate(property_list_sorted):
                value = i.properties.get(j)
                if value is not None:
                    reaction_sheet.write_string(x+1, y, text_type(value))
            reaction_sheet.write_string(
                x+1, len(property_list_sorted),
                text_type(i.id in model_reactions))
            reaction_sheet.write_string(
                x+1, len(property_list_sorted)+1,
                translated_equation)
            assoc = None
            if i.genes is None:
                continue
            elif isinstance(i.genes, string_types):
                assoc = boolean.Expression(i.genes)
                for j in assoc.variables:
                    gene_rxn[str(j)].append(i.id)
            else:
                variables = [boolean.Variable(g) for g in i.genes]
                assoc = boolean.Expression(boolean.And(*variables))
                for j in assoc.variables:
                    gene_rxn[str(j)].append(i.id)

        compound_sheet = workbook.add_worksheet(name='Compounds')

        compound_set = set()
        for compound in model.compounds:
            compound_set.update(compound.properties)

        compound_list_sorted = sorted(compound_set,
                                      key=lambda x: (x != 'id',
                                                     x != 'name', x))

        metabolic_model = self._model.create_metabolic_model()
        model_compounds = set(x.name for x in metabolic_model.compounds)
        for z, i in enumerate(compound_list_sorted + ['in_model']):
            compound_sheet.write_string(0, z, text_type(i))
        for x, i in enumerate(model.compounds):
            for y, j in enumerate(compound_list_sorted):
                value = i.properties.get(j)
                if value is not None:
                    compound_sheet.write_string(x+1, y, text_type(value))
            compound_sheet.write_string(
                x+1, len(compound_list_sorted),
                text_type(i.id in model_compounds))

        gene_sheet = workbook.add_worksheet(name='Genes')
        gene_sheet.write_string(0, 0, 'Gene')
        gene_sheet.write_string(0, 1, 'Reaction_in_model')
        gene_sheet.write_string(0, 2, 'Reaction_not_in_model')
        for x, i in enumerate(sorted(gene_rxn)):
            gene_sheet.write_string(x+1, 0, i)
            in_model = []
            not_in_model = []
            for r in gene_rxn.get(i):
                if r in model_reactions:
                    in_model.append(r)
                else:
                    not_in_model.append(r)
            gene_sheet.write_string(x+1, 1, '#'.join(in_model))
            gene_sheet.write_string(x+1, 2, '#'.join(not_in_model))

        exchange_sheet = workbook.add_worksheet(name='Exchange')

        exchange_sheet.write_string(0, 0, 'Compound ID')
        exchange_sheet.write_string(0, 1, 'Reaction ID')
        exchange_sheet.write_string(0, 2, 'Lower Limit')
        exchange_sheet.write_string(0, 3, 'Upper Limit')

        default_flux = model.default_flux_limit

        for x, (compound, reaction, lower, upper) in enumerate(
                itervalues(model.exchange)):
            if lower is None:
                lower = -default_flux

            if upper is None:
                upper = default_flux

            exchange_sheet.write(x+1, 0, text_type(compound))
            exchange_sheet.write(x+1, 1, text_type(reaction))
            exchange_sheet.write(x+1, 2, text_type(lower))
            exchange_sheet.write(x+1, 3, text_type(upper))

        limits_sheet = workbook.add_worksheet(name='Limits')

        limits_sheet.write_string(0, 0, 'Reaction ID')
        limits_sheet.write_string(0, 1, 'Lower Limit')
        limits_sheet.write_string(0, 2, 'Upper Limit')

        for x, limit in enumerate(itervalues(model.limits)):
            reaction_id, lower, upper = limit
            if lower is None:
                lower = -default_flux

            if upper is None:
                upper = default_flux

            limits_sheet.write(x+1, 0, reaction_id)
            limits_sheet.write(x+1, 1, lower)
            limits_sheet.write(x+1, 2, upper)

        workbook.close()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2018-2020  Jing Wang <wjingsjtu@gmail.com>

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import open
from builtins import super
import csv
import logging
from psamm.datasource.native import NativeModel
from psamm.importer import write_yaml_model
from psamm.util import mkdir_p
from future import standard_library
standard_library.install_aliases()  # noqa


def read_mapping(mapping):
    with open(mapping) as mapping_file:
        mapping_id = {}
        for row in csv.reader(mapping_file, delimiter='\t'):
            if row[1] == '':
                continue
            else:
                mapping_id[row[0]] = row[1]
    return(mapping_id)


def translate_id(id, mapping):
    if id not in list(mapping):
        return id
    return mapping[id]


class TranslatedModel(NativeModel):
    """A :class:`NativeModel` with translated ids based on reference model.

    The compound_map, reaction_map and compartment_map are three `dict`
    that use original id as key and new id as value.
    Use :meth:`.write_model` to output yaml files.
    """

    def __init__(self, ref_model,
                 compound_map, reaction_map, compartment_map=None):
        super().__init__()
        self._cpd_mapping_id = compound_map
        self._rxn_mapping_id = reaction_map
        if compartment_map is None:
            self._compartment_mapping_id = dict()
        else:
            self._compartment_mapping_id = compartment_map

        # initial a logger
        self._logger = logging.getLogger(__name__)

        # inherit properties
        self._properties = ref_model._properties
        self.version_string = 'model mapped from: %s' % self.version_string
        self.biomass_reaction = translate_id(
            ref_model.biomass_reaction, self.rxn_mapping_id)

        # translate ids
        self._translate_compartment(ref_model)
        self._translate_compounds(ref_model)
        self._translate_reactions(ref_model)
        self._translate_exchange(ref_model)
        self._translate_limits(ref_model)

    @property
    def compartment_mapping_id(self):
        return self._compartment_mapping_id

    @property
    def cpd_mapping_id(self):
        return self._cpd_mapping_id

    @property
    def rxn_mapping_id(self):
        return self._rxn_mapping_id

    def _translate_compartment(self, ref_model):
        """Translate compartment ids."""
        # change compartments
        for comp in ref_model.compartments:
            new_id = translate_id(comp.id, self.compartment_mapping_id)
            comp.properties['id'] = new_id
            comp.properties['original_id'] = [comp.id]
            comp = comp.__class__(comp.properties, filemark=comp.filemark)
            self.compartments.add_entry(comp)
        # change comparment_boundaries
        for c1, c2 in ref_model.compartment_boundaries:
            c1 = translate_id(c1, self.compartment_mapping_id)
            c2 = translate_id(c2, self.compartment_mapping_id)
            self.compartment_boundaries.add((c1, c2))
        self.extracellular_compartment = translate_id(
            ref_model.extracellular_compartment, self.compartment_mapping_id)

    def _translate_compounds(self, ref_model):
        """Translate compound ids and corresponding exchange reactions."""
        for cpd in ref_model.compounds:
            new_id = translate_id(cpd.id, self.cpd_mapping_id)
            if new_id not in self.compounds:  # do not duplicate compounds
                cpd.properties['id'] = new_id
                cpd.properties['original_id'] = [cpd.id]
                # refresh with changed properties
                cpd = cpd.__class__(cpd.properties, filemark=cpd.filemark)
                self.compounds.add_entry(cpd)
            else:
                self.compounds[new_id].properties['original_id'].append(cpd.id)
                self._logger.warning(
                    'Multiple compounds: %s will be translated to %s',
                    self.compounds[new_id].properties['original_id'],
                    new_id)

    def _compound_trans(self, id):
        """Make a directly callable method for compound.translate()"""
        return translate_id(id, self.cpd_mapping_id)

    def _translate_reactions(self, ref_model):
        """Translate reaction ids and equations."""

        for rxn in ref_model.reactions:
            new_id = translate_id(rxn.id, self.rxn_mapping_id)
            if new_id not in self.reactions:  # do not duplicate reactions
                rxn.properties['id'] = new_id
                rxn.properties['original_id'] = rxn.id
                compounds = list()
                for cpd, v in rxn.equation.compounds:
                    cpd = cpd.translate(self._compound_trans)
                    cpd = cpd.in_compartment(
                        translate_id(cpd.compartment,
                                     self.compartment_mapping_id)
                    )
                    compounds.append((cpd, v))
                rxn.equation = rxn.equation.__class__(
                    rxn.equation.direction, compounds
                )
                # test whether the same compound occurs at both sides
                compound_left = set((c for c, _ in rxn.equation.left))
                compound_right = set((c for c, _ in rxn.equation.right))
                compound_common = compound_left.intersection(compound_right)
                if len(compound_common) > 0:
                    self._logger.error((
                        'Reaction %s will have '
                        'the same compounds %s on both sides: %s'),
                        rxn.id, compound_common, rxn.equation)
                # refresh with changed properties
                rxn = rxn.__class__(rxn.properties, filemark=rxn.filemark)
                self.reactions.add_entry(rxn)
            else:
                self._logger.error(
                    ('Reaction %s is omitted because its new id %s has been '
                     'assigned to %s'),
                    rxn.id,
                    new_id,
                    self.reactions[new_id].properties['original_id'])

    def _translate_exchange(self, ref_model):
        """Translate exchange reactions."""
        # cpd is :class:`Compound(name, compartment)`
        for cpd in ref_model.exchange:
            new_cpd = cpd.translate(self._compound_trans)
            new_cpd = new_cpd.in_compartment(self.extracellular_compartment)
            if new_cpd not in self.exchange:  # do not duplicate exchange
                lower, upper = ref_model.exchange[cpd][2:]
                name = 'EX_%s(%s)' % (new_cpd.name,
                                      self.extracellular_compartment)
                self.exchange[new_cpd] = (new_cpd, name, lower, upper)
            else:
                self._logger.error(
                    ('Exchange reaction of compound %s is omitted because '
                     'its new id %s has been assigned multiple compounds: %s'),
                    cpd.name,
                    new_cpd.name,
                    self.compounds[new_cpd.name].properties['original_id'])

    def _translate_limits(self, ref_model):
        """"Translate reaction limits."""
        for rxn in ref_model.limits:
            new_rxn = translate_id(rxn, self.rxn_mapping_id)
            if new_rxn not in self.limits:  # do not duplicate limits
                lower, upper = ref_model.limits[rxn][1:]
                self.limits[new_rxn] = (new_rxn, lower, upper)
            else:
                self._logger.error(
                    ('Limits of reaction %s is omitted because its new id %s '
                     'has been assigned to %s'),
                    rxn,
                    new_rxn,
                    self.reactions[new_rxn].properties['original_id'])

    def write_model(self, dest, **kwargs):
        """Output model into YAML files."""
        mkdir_p(dest)
        write_yaml_model(self, dest, **kwargs)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019-2022  Christopher Powers <c-11060@uri.edu>
# Copyright 2021-2022  Jason Vailionis <jason_vailionis@uri.edu>

from __future__ import unicode_literals
import logging
import argparse
import yaml
import os
import sys
from collections import defaultdict
from collections import OrderedDict
import re
from six import iteritems
from psamm.datasource.native import parse_reaction_equation_string
from psamm.datasource.native import ModelReader
from psamm.datasource.entry import (DictReactionEntry as ReactionEntry)
from psamm.datasource.context import FileMark
from psamm.datasource.reaction import Reaction, Compound, Direction
from psamm.formula import Formula, ParseError
from pkg_resources import resource_filename
import pkg_resources
from psamm.command import _trim
from six import add_metaclass, iteritems, text_type
import abc
from urllib.error import HTTPError
logger = logging.getLogger(__name__)
if sys.version_info.minor > 5:
    try:
        from Bio.KEGG import REST
    except ImportError:
        logger.warning("WARNING: Biopython package not found! "
                       "Some functions will be unusable")
    try:
        from libchebipy._chebi_entity import ChebiEntity
    except ImportError:
        logger.warning("WARNING: The Chebi API package not found! "
                       "Some functions will be unusable")


class ParseError2(Exception):
    """Exception used to signal errors while parsing"""


class VersionError(Exception):
    """Exception used to signal incorrect python version"""


class CommandError(Exception):
    """Error from running a command.

    This should be raised from a ``Command.run()`` if any arguments are
    misspecified. When the command is run and the ``CommandError`` is raised,
    the caller will exit with an error code and print appropriate usage
    information.
    """


class InputError(Exception):
    """Exception used to signal a general input error. Exiting..."""


def dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


@add_metaclass(abc.ABCMeta)
class Command(object):
    """Represents a command in the interface, operating on a model.

    The constructor will be given the NativeModel and the command
    linenamespace. The subclass must implement :meth:`run` to
    handle commandexecution. The doc string will be used as
    documentation for the command in the command line interface.

    In addition, :meth:`init_parser` can be implemented as a
    classmethod which will allow the command to initialize an
    instance of :class:`argparse.ArgumentParser` as desired.
    The resulting argumentnamespace will be passed to the
    constructor.
    """

    def __init__(self, args):
        self._args = args

    @classmethod
    def init_parser(cls, parser):
        """Initialize command line parser (:class:`argparse.ArgumentParser`)"""

    @abc.abstractmethod
    def run(self):
        """Execute command"""

    def argument_error(self, msg):
        """Raise error indicating error parsing an argument."""
        raise CommandError(msg)

    def fail(self, msg, exc=None):
        """Exit command as a result of a failure."""
        logger.error(msg)
        if exc is not None:
            logger.debug('Command failure caused by exception!', exc_info=exc)
        sys.exit(1)


def parse_orthology(orthology, type, col):
    '''
    function to parse orthology tables. The default format for these
    tables is the default eggnog output, but custom colummn numers can be
    passed through the --col argument
    '''
    # Dictionary of reactions to genes
    asso_dict = defaultdict(lambda: [])
    # Populate the dictionary
    with open(orthology, "r") as infile:
        for line in infile:
            if line.startswith('#'):
                continue
            line = line.rstrip()
            listall = re.split("\t", line)
            # Add handling for a provided column number.
            if col:
                if len(listall[col-1]) > 0:
                    keys = listall[col-1].split(',')
                else:
                    keys = []
            else:
                if type == "R":
                    keys = listall[14].split(',')
                elif type == "EC":
                    keys = listall[10].split(',')
                elif type == "KO":
                    keys = listall[11].split(',')
                elif type == "tcdb":
                    if listall[17] == '-':
                        continue
                    keys = listall[17].split(',')
            if len(keys) == 0:
                continue
            for k in keys:
                asso_dict[k].append(listall[0])
    return(asso_dict)


def model_reactions(reaction_entry_list):
    '''
    Function to sort the downloaded kegg object into a format
    that is compatible with the psamm api for storage in
    a reactions.yaml file.
    '''
    for reaction in reaction_entry_list:

        d = OrderedDict()
        d['id'] = reaction.id

        enzymes_list = []
        for i in reaction.enzymes:
            enzymes_list.append(i)

        pathways_list = []
        if reaction.pathways is not None:
            for i in reaction.pathways:
                pathways_list.append(i[1])
        if len(pathways_list) == 0:
            pathways_list = None

        orth_list = []
        for i in reaction.orthology:
            orth_list.append(i)
        if len(orth_list) == 0:
            orth_list = None

        if hasattr(reaction, 'name') and reaction.name is not None:
            d['name'] = reaction.name
        if hasattr(reaction, 'names') and reaction.names is not None:
            names_l = []
            for i in reaction.names:
                names_l.append(i)
            d['names'] = names_l
        if hasattr(reaction, 'equation') and reaction.equation is not None:
            d['equation'] = str(reaction.equation)
        if hasattr(reaction, 'definition') and reaction.definition is not None:
            d['KEGG_definition'] = reaction.definition
        if hasattr(reaction, 'enzymes') and reaction.enzymes is not None:
            d['enzymes'] = enzymes_list
        if hasattr(reaction, 'pathways') and reaction.pathways is not None:
            d['pathways'] = pathways_list
        if hasattr(reaction, 'comment') and reaction.comment is not None:
            d['comment'] = str(reaction.comment)
        if hasattr(reaction, 'tcdb_family') and \
                reaction.tcdb_family is not None:
            d['tcdb_family'] = str(reaction.tcdb_family)
        if hasattr(reaction, 'substrates') and \
                reaction.substrates is not None:
            d['substrates'] = reaction.substrates
        if hasattr(reaction, 'genes') and \
                reaction.genes is not None:
            d['genes'] = str(reaction.genes)
        if hasattr(reaction, 'orthology') and \
                reaction.orthology is not None:
            d['orthology'] = orth_list
        yield d


def clean_reaction_equations(reaction_entry_list):
    '''
    This function handles specific edge cases in the kegg
    format of reaction equations that are incompatible with
    the psamm format. Takes the downloaded dictionary of reactions
    and returns the same dictionary with modified equations,
    if necessary.
    '''
    generic = []
    for reaction in reaction_entry_list:
        equation = re.sub(r'\(.*?\)', lambda x: ''.join(x.group(0).split()),
                          str(reaction.equation))
        equation_out = []
        comp = re.split(" ", equation)
        comp = comp + [""]

        for i in range(0, len(comp)-1):
            # Handles unusual stoichiometry at the beginning of compounds
            if "(" in comp[i] and "C" in comp[i+1]:
                if i > 0:
                    if "=" in comp[i-1] or "+" in comp[i-1]:
                        equation_out.append(comp[i])
                        continue
                else:
                    equation_out.append(comp[i])
                    continue
                generic.append(reaction.id)
            # special handling to retain stoichiometry of (n+1) and (n-1), etc.
            # at the end of the compound
            if "(" in comp[i]:  # and (not "(n+" in i or not "(n-" in i):
                comp[i] = comp[i].replace("(", "[")
                comp[i] = comp[i].replace(")", "]")
                generic.append(reaction.id)
            equation_out.append(comp[i])
        equation = ' '.join(equation_out)
        reaction.__dict__['values']['equation'] = [equation]

    return(reaction_entry_list, generic)


def create_model_api(out, rxn_mapping, verbose, use_rhea, default_compartment):
    '''
    This function creates a draft model based on the reactions:genes
    file specified in the rxn variable. This function generates
    reaction and compound information by utilizing the kegg
    REST api to download the reaction information and uses psamm
    functions to parse out the kegg data.
    '''
    with open(os.path.join(out, "log.tsv"), "a+") as f:
        f.write("List of invalid Kegg IDs: \n")

    if verbose:
        logger.info("There are {} reactions to download"
                    ".".format(len(rxn_mapping)))
    # Generate the yaml file for the reactions
    reaction_entry_list = []
    count = 0
    for entry in _download_kegg_entries(out, rxn_mapping, None, ReactionEntry):
        reaction_entry_list.append(entry)
        count += 1
        if verbose and count % 25 == 0:
            logger.info("{}/{} reactions downloaded..."
                        .format(count, len(rxn_mapping)))

    # Clean up the kegg reaction dict.
    reaction_entry_list, gen = clean_reaction_equations(reaction_entry_list)

    # Sets up the yaml object for the reactions and writes
    # out the parsed reaction information to a reactions.yaml
    # file
    yaml.add_representer(OrderedDict, dict_representer)
    yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                         dict_constructor)
    yaml_args = {'default_flow_style': False,
                 'encoding': 'utf-8',
                 'allow_unicode': True}

    # Load database once if --rhea is given
    if use_rhea:
        rhea_db = RheaDb(resource_filename('psamm',
                         'external-data/chebi_pH7_3_mapping.tsv'))
    else:
        rhea_db = None

    compound_entry_list = []
    compound_set = set()
    for reaction in reaction_entry_list:
        eq = parse_reaction_equation_string(reaction.equation, 'c')
        for i in eq.compounds:
            compound_set.add(str(i[0].name))
    if verbose:
        logger.info("There are {} compounds to download."
                    .format(len(compound_set)))
    count = 0
    logger.info("Downloading kegg entries and performing charge correction")
    for entry in _download_kegg_entries(out, compound_set,
                                        rhea_db, CompoundEntry):
        compound_entry_list.append(entry)
        count += 1
        if verbose and count % 25 == 0:
            logger.info("{}/{} compounds downloaded...".
                        format(count, len(compound_set)))
    with open(os.path.join(out, 'compounds.yaml'), 'w+') as f:
        yaml.dump(list(model_compounds(compound_entry_list)), f, **yaml_args)

    # generate a yaml file for all generic compounds
    generic_compounds_list = check_generic_compounds(compound_entry_list)
    generic_entry_list = []
    logger.info("Downloading kegg entries for generic compounds and "
                "performing charge correction")
    for entry in _download_kegg_entries(out, generic_compounds_list, rhea_db,
                                        CompoundEntry):
        generic_entry_list.append(entry)
    with open(os.path.join(out, 'compounds_generic.yaml'), 'w+') as f:
        yaml.dump(list(model_generic_compounds(generic_entry_list)),
                  f, **yaml_args)
    with open(os.path.join(out, 'log.tsv'), "a+") as f:
        f.write("\nThere are {} generic compounds in the model\n".format(str(
            len(generic_compounds_list))))
        f.write("Generic compounds:\n")
        for i in generic_entry_list:
            f.write("{}".format(i.id))
            if i.name:
                f.write("|{}".format(i.name))
            else:
                f.write("|")
            if i.formula:
                f.write("|{}\n".format(i.formula))
            else:
                f.write("|\n")

    # generate a yaml for the reactions containing generic compounds and
    # and reactions without without generic compounds
    reaction_list_out = []
    reaction_list_generic = []
    compound_list_out = set()

    with open(os.path.join(out, 'log.tsv'), 'a+') as f:
        f.write("\nThe reactions containing these generic compounds are: \n")
        for reaction in reaction_entry_list:
            if any(i in str(reaction.equation)
                   for i in generic_compounds_list) or reaction.id in gen:
                f.write("{}".format(reaction.id))
                if reaction.name:
                    f.write("|{}".format(reaction.name))
                else:
                    f.write("|")
                if reaction.equation:
                    f.write("|{}\n".format(reaction.equation))
                else:
                    f.write("|\n")
                reaction_list_generic.append(reaction)
            else:
                reaction_list_out.append(reaction)
                for c in str(reaction.equation):
                    compound_list_out.add(str(c))

    with open(os.path.join(out, 'reactions.yaml'), 'w+') as f:
        yaml.dump(list(model_reactions(reaction_list_out)), f, **yaml_args)
    with open(os.path.join(out, 'reactions_generic.yaml'), 'w+') as f:
        yaml.dump(list(model_reactions(reaction_list_generic)), f, **yaml_args)

    # Create a model.yaml file
    with open('{}/model.yaml'.format(out), mode='w') as myam:
        myam.write('default_flux_limit: 100\n')
        myam.write('default_compartment: {}\n'.format(default_compartment))
        myam.write('extracellular: null\n')
        myam.write('biomass: null\n')
        myam.write('compounds:\n')
        myam.write('- include: ./compounds.yaml\n')
        myam.write('reactions:\n')
        myam.write('- include: ./reactions.yaml\n')
        myam.write('- include: ./gene-association.tsv\n')
        myam.write('  format: tsv\n')
        myam.write('model:\n')
        myam.write('- include: model_def.tsv\n')

    with open('{}/gene-association.tsv'.format(out), mode='w') as outfile:
        outfile.write('id\tgenes\n')
        for reaction in reaction_list_out:
            if len(rxn_mapping[reaction.id]) == 1:
                gene_asso = rxn_mapping[reaction.id]
            else:
                gene_asso = ['({})'.format(gene)
                             for gene in rxn_mapping[reaction.id]]
            outfile.write('{}\t{}\n'.format(reaction.id,
                          ' or '.join(gene_asso)))

    with open('{}/gene-association_generic.tsv'.format(out), mode='w') \
            as outfile:
        outfile.write('id\tgenes\n')
        for reaction in reaction_list_generic:
            if len(rxn_mapping[reaction.id]) == 1:
                gene_asso = rxn_mapping[reaction.id]
            else:
                gene_asso = ['({})'.format(gene)
                             for gene in rxn_mapping[reaction.id]]
            outfile.write('{}\t{}\n'.format(reaction.id,
                          ' or '.join(gene_asso)))

    # Create a model_def file with all of the new reactions
    with open('{}/model_def.tsv'.format(out), mode='w') as f:
        for reaction in reaction_list_out:
            f.write('{}\n'.format(reaction.id))

    # Write out some final statistics for the model
    with open(os.path.join(out, 'log.tsv'), 'a+') as f:
        f.write("\nThere are {} reactions in the model".format(str(
                len(reaction_list_out))))
        f.write("\nThere are {} compounds in the model\n".format(str(
                len(compound_list_out))))
    if verbose:
        logger.info("\nThere are {} reactions in the model".format(str(
                    len(reaction_list_out))))
        logger.info("\nThere are {} compounds in the model\n".format(str(
                    len(compound_entry_list))))


def model_compounds(compound_entry_list):
    '''
    Function to sort the downloaded	kegg object into a format
    that is	compatible with the psamm api for storage in
    a compounds.yaml file.
    '''
    non_gen_compounds = []
    for compound in compound_entry_list:
        try:
            form = Formula.parse(str(compound.formula))
            if form.is_variable():
                continue
            elif compound.formula is None:
                continue
            elif 'R' in str(compound.formula):
                continue
            else:
                d = OrderedDict()
                d['id'] = compound.id
                non_gen_compounds.append(compound.id)
                if hasattr(compound, 'name') and compound.name is not None:
                    d['name'] = compound.name
                if hasattr(compound, 'names') and \
                        compound.names is not None:
                    names_l = []
                    for i in compound.names:
                        names_l.append(i)
                    d['names'] = names_l
                if hasattr(compound, 'formula') and \
                        compound.formula is not None:
                    d['formula'] = str(compound.formula)
                if hasattr(compound, 'mol_weight') and \
                        compound.mol_weight is not None:
                    d['mol_weight'] = compound.mol_weight
                if hasattr(compound, 'comment') and \
                        compound.comment is not None:
                    d['comment'] = str(compound.comment)
                if hasattr(compound, 'dblinks') and \
                        compound.dblinks is not None:
                    for key, value in compound.dblinks:
                        if key != 'ChEBI':
                            d['{}'.format(key)] = value
                if hasattr(compound, 'chebi') \
                        and compound.chebi is not None:
                    d['ChEBI'] = compound.chebi
                if hasattr(compound, 'chebi_all') \
                        and compound.chebi_all is not None:
                    d['ChEBI_all'] = compound.chebi_all
                if hasattr(compound, 'charge') \
                        and compound.charge is not None:
                    d['charge'] = compound.charge
                yield d
        except ParseError:
            logger.warning("import of {} failed"
                           " and will not be imported into compounds.yaml "
                           "or compounds_generic.yaml".format(compound.id))
            continue


def check_generic_compounds(compound_entry_list):
    '''
    Function for checking if the compound formulation is
    compatible with psamm. generalized rules for this are
    that compounds must have a formula, the formula cannot
    be variable (e.g. presence of X), and R groups are
    generally discouraged.
    '''
    generic_compounds_list = []
    for compound in compound_entry_list:
        try:
            form = Formula.parse(str(compound.formula))
            if form.is_variable():
                generic_compounds_list.append(compound.id)
            elif compound.formula is None:
                generic_compounds_list.append(compound.id)
            elif 'R' in str(compound.formula):
                generic_compounds_list.append(compound.id)
            else:
                continue
        except ParseError:
            generic_compounds_list.append(compound.id)
    return(generic_compounds_list)


def model_generic_compounds(compound_entry_list):
    '''
    Function to sort the downloaded	kegg object into a format
    that is	compatible with the psamm api for storage in
    a generic_compounds.yaml file. This function contains
    special error handling for improperly formatted compounds
    '''
    non_gen_compounds = []
    for compound in compound_entry_list:
        try:
            d = OrderedDict()
            d['id'] = compound.id
            non_gen_compounds.append(compound.id)
            if hasattr(compound, 'name') and compound.name is not None:
                d['name'] = compound.name
            if hasattr(compound, 'names') and compound.names is not None:
                names_l = []
                for i in compound.names:
                    names_l.append(i)
                d['names'] = names_l
            if hasattr(compound, 'formula') and compound.formula is not None:
                d['formula'] = str(compound.formula)
            if hasattr(compound, 'mol_weight') and \
                    compound.mol_weight is not None:
                d['mol_weight'] = compound.mol_weight
            if hasattr(compound, 'comment') and compound.comment is not None:
                d['comment'] = str(compound.comment)
            if hasattr(compound, 'dblinks') and compound.dblinks is not None:
                for key, value in compound.dblinks:
                    if key != 'ChEBI':
                        d['{}'.format(key)] = value
            if hasattr(compound, 'chebi') and compound.chebi is not None:
                d['ChEBI'] = compound.chebi
            if hasattr(compound, 'chebi_all') and \
                    compound.chebi_all is not None:
                d['ChEBI_all'] = compound.chebi_all
            if hasattr(compound, 'charge') and compound.charge is not None:
                d['charge'] = compound.charge
            yield d
        except ParseError:
            logger.warning("{} is improperly formatted "
                           " and will not be imported into "
                           "compounds_generic.yaml".format(compound.id))
            continue


def _download_kegg_entries(out, rxn_mapping, rhea, entry_class, context=None):
    '''
    Downloads the kegg entry associated with a reaction or
    a compound and stores each line in an object that can
    be parsed as a reaction or a compound, depending on the
    input
    '''
    # go through the rxn mapping dict and pull all of the
    # reaction information out of the Kegg API
    with open(os.path.join(out, 'log.tsv'), "a+") as f:
        for reactions in rxn_mapping:
            # Check for standard formatting of R and C. These can be treated
            # the same way
            if reactions[0] == "R" or reactions[0] == "C":
                # Try except loop to catch a failure to download.
                # On a failure, the failure gets recorded in the log
                try:
                    request = REST.kegg_get(reactions)
                except HTTPError:
                    f.write("".join(["  - ", reactions, "\n"]))
                    continue
                entry_line = None
                section_id = None
                reaction = {}
                for lineno, line in enumerate(request):
                    line = line.rstrip()
                    if line == "///":
                        continue
                    if entry_line is None:
                        entry_line = lineno
                    # Look for the beginning of the section
                    m = re.match(r'([A-Z_]+)\s+(.*)', line.rstrip())
                    if m is not None:
                        section_id = m.group(1).lower()
                        reaction[section_id] = [m.group(2)]
                    elif section_id is not None:
                        reaction[section_id].append(line.strip())
                    else:
                        raise ParseError2(
                            'Missing section identifier at line \
                            {}'.format(lineno))
                mark = FileMark(context, entry_line, 0)
                yield entry_class(reaction, rhea, filemark=mark)


def parse_rxns_from_EC(rxn_mapping, out, verbose):
    """
    Functions converts gene associations to EC into gene
    associations for reaction IDs. Returns a dictionary
    of Reaction IDs to genes.
    """
    rxn_dict = defaultdict(lambda: [])
    if verbose:
        logger.info("Downloading reactions associated with EC...")
        logger.info("There are {} ECs download".format(len(rxn_mapping)))
        count = 0
    with open(os.path.join(out, "log.tsv"), "a+") as f:
        for reactions in rxn_mapping:
            if verbose:
                if count % 25 == 0:
                    logger.info("{}/{} have been downloaded".format(count,
                                len(rxn_mapping)))
                count += 1
            try:
                request = REST.kegg_get(reactions)
            except HTTPError:
                f.write("".join(["  - ", reactions, "\n"]))
                continue
            entry_line = None
            section_id = None
            reaction = {}
            for lineno, line in enumerate(request):
                line = line.rstrip()
                if line == "///":
                    continue
                if entry_line is None:
                    entry_line = lineno
                # Look for the beginning of the section
                m = re.match(r'([A-Z_]+)\s+(.*)', line.rstrip())
                if m is not None:
                    section_id = m.group(1).lower()
                    reaction[section_id] = [m.group(2)]
                elif section_id is not None:
                    reaction[section_id].append(line.strip())
                else:
                    raise ParseError2('Missing section identifier at line '
                                      '{}'.format(lineno))
            if "all_reac" in reaction:
                listall = re.split(" ", reaction["all_reac"][0])
                for r in listall:
                    if r[0] == "R":
                        rxn_dict[r] += rxn_mapping[reactions]

    return(rxn_dict)


def parse_rxns_from_KO(rxn_mapping, out, verbose):
    """
    Functions converts gene associations to EC into gene
    associations for reaction IDs. Returns a dictionary
    of Reaction IDs to genes.
    """
    rxn_dict = defaultdict(lambda: [])
    if verbose:
        logger.info("Downloading reactions associated with KO...")
        logger.info("There are {} KOs download".format(len(rxn_mapping)))
        count = 0
    with open(os.path.join(out, "log.tsv"), "a+") as f:
        for reactions in rxn_mapping:
            if verbose:
                if count % 25 == 0:
                    logger.info("{}/{} have been downloaded".format(count,
                                len(rxn_mapping)))
                count += 1
            try:
                request = REST.kegg_get(reactions)
            except HTTPError:
                f.write("".join(["  - ", reactions, "\n"]))
                continue
            entry_line = None
            section_id = None
            reaction = {}
            for lineno, line in enumerate(request):
                line = line.rstrip()
                if line == "///":
                    continue
                if entry_line is None:
                    entry_line = lineno
                # Look for the beginning of the section
                m = re.match(r'([A-Z_]+)\s+(.*)', line.rstrip())
                if m is not None:
                    section_id = m.group(1).lower()
                    reaction[section_id] = [m.group(2)]
                elif section_id is not None:
                    reaction[section_id].append(line.strip())
                else:
                    raise ParseError2(
                        'Missing section identifier at line \
                        {}'.format(lineno))
            if "dblinks" in reaction:
                for i in reaction["dblinks"]:
                    if i[0:2] == "RN":
                        listall = re.split(" ", i[1:])
                        for r in listall:
                            if r[0] == "R":
                                rxn_dict[r] += rxn_mapping[reactions]
    return(rxn_dict)


def gen_transporters(path, gene_asso, tp_sub_dict, tp_fam_dict, chebi_dict,
                     compartment_in, compartment_out):
    with open(os.path.join(path, "transporter_log.tsv"), 'w') \
            as log:
        log.write("compounds for which transporters were predicted, but"
                  " are not found in the model:\n")
        eq = defaultdict(lambda: [])
        for id, genes in gene_asso.items():
            # if there is just one substrate and it is in the porin family
            # assume passive diffusion/porin behavior
            if len(tp_sub_dict[id]) == 1:  # id[0] == "1" and
                for cpd in tp_sub_dict[id]:
                    if cpd in chebi_dict:
                        eq[id].append(("{}_diff".format(chebi_dict[cpd][0]),
                                       Reaction(Direction.Both,
                                       {Compound(chebi_dict[cpd][0]).
                                        in_compartment(compartment_in): -1,
                                        Compound(chebi_dict[cpd][0]).
                                        in_compartment(compartment_out): 1})))
                    else:
                        log.write("{}\t{}\n".format("Compound not in "
                                  "model: ", cpd, id))
            # Otherwise, create a template reaction entry with no
            # formulation for user curation
            else:
                eq[id].append(("TP_{}".format(id),
                              Reaction(Direction.Both,
                                       {Compound("").in_compartment(
                                        compartment_in): -1,
                                        Compound("").in_compartment(
                                        compartment_out): 1})))

        # Sets up the yaml object for the reactions and writes
        # out the parsed reaction information to a reactions.yaml
        # file
        yaml.add_representer(OrderedDict, dict_representer)
        yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                             dict_constructor)
        yaml_args = {'default_flow_style': False,
                     'encoding': 'utf-8',
                     'allow_unicode': True}

        # construct a new reactions.yaml file based on this.
        reaction_entry_list = []
        mark = FileMark(None, 0, 0)
        rhea = None
        with open(os.path.join(path, "transporters.yaml"),
                  "w") as f:
            for tcdb in eq:
                for rxn in eq[tcdb]:
                    fam_id = ".".join(tcdb.split(".")[0:3])
                    reaction = {'transport_name': rxn[0], 'entry': [rxn[0]],
                                'name': [tp_fam_dict[fam_id]],
                                'equation': [str(rxn[1])],
                                'enzyme': [tcdb],
                                'tcdb_family': tp_fam_dict[fam_id],
                                'substrates': tp_sub_dict[tcdb],
                                'genes': "({})".format(" or ".join(
                                    gene_asso[tcdb]))}
                    entry = ReactionEntry(reaction, rhea, filemark=mark)
                    reaction_entry_list.append(entry)

            yaml.dump(list(model_reactions(reaction_entry_list)), f,
                      **yaml_args)


class main_transporterCommand(Command):
    """Predicts the function of transporter reactions.

    Output based on algorithm outlined in PMID: 18586723"""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument('--annotations', metavar='path',
                            help='Path to the annotation file from eggnog')
        parser.add_argument('--db_substrates', type=str,
                            help='OPTIONAL. Path to a custom substrate file.')
        parser.add_argument('--db_families', type=str,
                            help='OPTIONAL. Path to a custom family file.')
        parser.add_argument('--model', metavar='path',
                            help='Path to the model directory')
        parser.add_argument('--compartment_in', default='c',
                            help='abbreviation for the internal compartment')
        parser.add_argument('--compartment_out', default='e',
                            help='abbreviation for the external compartment')
        parser.add_argument('--col', default=None, help='If providing your own'
                            ' annotation table, specify column for the R, EC, '
                            'or KO number. The default is to parse eggnog'
                            ' output.', type=int)
        super(main_transporterCommand, cls).init_parser(parser)

    def run(self):
        if sys.version_info.minor < 6:
            raise VersionError("Biopython only compatible with python > 3.5.")
        '''Entry point for the transporter assignment'''

        # Check the validity of the input values
        if not self._args.annotations:
            raise InputError('Please specify a path to the annotations')
        if self._args.db_substrates:
            logger.info("Using custom transporter families")
            substrate = self._args.db_substrates
        else:
            logger.info("Using the default transporter families from TCDB. "
                        "Downloaded from: ")
            logger.info("http://www.tcdb.org/cgi-bin/projectv/public/"
                        "getSubstrates.py")
            substrate = resource_filename('psamm',
                                          'external-data/tcdb_substrates.tsv')
        if self._args.db_families:
            logger.info("Using custom transporter families")
            family = self._args.db_families
        else:
            logger.info("Using the default transporter families from TCDB. "
                        "Downloaded from: ")
            logger.info("http://www.tcdb.org/cgi-bin/projectv/public/"
                        "families.py")
            family = resource_filename('psamm',
                                       'external-data/tcdb_families.tsv')
        if not self._args.model:
            raise InputError('Please specify a directory to an existing model')

        # read in the model and build a dictionary of Chebi IDs
        mr = ModelReader.reader_from_path(self._args.model)
        nm = mr.create_model()

        chebi_dict = defaultdict(lambda: [])
        for cpd in nm.compounds:
            if 'ChEBI' in cpd.__dict__['_properties']:
                chebi = re.split(' ', cpd.__dict__['_properties']['ChEBI'])
                for i in chebi:
                    chebi_dict["CHEBI:{}".format(i)].append(cpd.id)

        # Read in the reaction substrates
        tp_sub_dict = defaultdict(lambda: [])
        with open(substrate, 'r', encoding="utf8") as infile:
            for line in infile:
                line = line.rstrip()
                listall = re.split("\t", line)
                substrates = re.split(r"\|", listall[1])
                sub_out = []
                for i in substrates:
                    sub_out.append(re.split(";", i)[0])
                tp_sub_dict[listall[0]] = sub_out

        # read in the reaction families
        tp_fam_dict = defaultdict(lambda: '')
        with open(family, 'r', encoding="utf8") as infile:
            for line in infile:
                line = line.rstrip()
                listall = re.split("\t", line)
                tp_fam_dict[listall[0]] = listall[1]

        # build a gene association dictionary
        # Launch the parse_orthology function to contruct a gene
        # association file
        gene_asso = parse_orthology(self._args.annotations, "tcdb",
                                    self._args.col)

        # Build the transporter databse based on what is in
        # the annotations
        if "model.yaml" in self._args.model:
            path = self._args.model.rstrip("model.yaml")
        else:
            path = self._args.model
        gen_transporters(path, gene_asso, tp_sub_dict, tp_fam_dict,
                         chebi_dict, self._args.compartment_in,
                         self._args.compartment_out)


class main_databaseCommand(Command):
    """Generate a database of compounds and reactions"""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument('--annotations', metavar='path',
                            help='Path to the annotation file from Eggnog')
        parser.add_argument('--type', type=str,
                            help='Define whether to build the model on'
                            ' reaction ID, KO, or EC.\noptions are: '
                            '[R, KO, EC]')
        parser.add_argument('--out', metavar='out',
                            help='Path to the output location for the model '
                            'directory. This will be created for you.')
        parser.add_argument('--col', default=None, help='If providing your '
                            'own annotation table, specify column for the R, '
                            'EC, or KO number. The default is to parse eggnog'
                            ' output.', type=int)
        parser.add_argument('--verbose', help='Report progress  verbose mode',
                            action='store_true')
        parser.add_argument('--default_compartment',
                            help='Define abbreviation for the default '
                            'compartment',
                            default='c')
        parser.add_argument('--rhea', help='Resolve protonation states '
                            'using major microspecies at pH 7.3 using '
                            'Rhea-ChEBI mappings', action='store_true')
        parser.add_argument('--force', help='force rewrite of existing '
                            'directories', action='store_true')
        super(main_databaseCommand, cls).init_parser(parser)

    def run(self):
        """Entry point for the database generation script"""
        # check if required packages are installed
        if 'Bio.KEGG.REST' not in sys.modules:
            quit('No biopython package found. '
                 'Please run <pip install biopython>''')
        if "libchebipy._chebi_entity" not in sys.modules:
            quit('The Chebi API is not installed. '
                 'Please run <pip install libchebipy>')
        if sys.version_info.minor < 6:
            raise VersionError("Biopython only compatible with python > 3.5.")

        # Check the validity of the input values
        if not self._args.annotations:
            raise InputError('Please specify a path to the eggnog annotations')
        if not self._args.type:
            raise InputError('Please specify one of R, KO, or EC as the type')
        if not self._args.out:
            raise InputError('Please specify an output directory')
        if os.path.exists(self._args.out) and not self._args.force:
            raise InputError('The output directory already exists! Exiting...')
            exit()
        elif os.path.exists(self._args.out) is False:
            os.mkdir(self._args.out)

        # Check the format of the eggnog annotation file.
        # Add some code here to check the input file.

        # Launch the parse_orthology function to contruct a
        # gene association file
        ortho_dict = parse_orthology(self._args.annotations, self._args.type,
                                     self._args.col)

        # convert EC to reactions
        if self._args.type == "EC":
            ortho_dict = parse_rxns_from_EC(ortho_dict, self._args.out,
                                            self._args.verbose)
        elif self._args.type == "KO":
            ortho_dict = parse_rxns_from_KO(ortho_dict, self._args.out,
                                            self._args.verbose)

        # Create the model using the kegg api
        create_model_api(self._args.out, ortho_dict, self._args.verbose,
                         self._args.rhea, self._args.default_compartment)


def main(command_class=None, args=None):
    """Run the command line interface with the given :class:`Command`.

    If no command class is specified the user will be able to select a specific
    command through the first command line argument. If the ``args`` are
    provided, these should be a list of strings that will be used instead of
    ``sys.argv[1:]``. This is mostly useful for testing.
    """

    # Set up logging for the command line interface
    if 'PSAMM_DEBUG' in os.environ:
        level = getattr(logging, os.environ['PSAMM_DEBUG'].upper(), None)
        if level is not None:
            logging.basicConfig(level=level)
    else:
        logging.basicConfig(
            level=logging.INFO, format='%(levelname)s: %(message)s')

    parser = argparse.ArgumentParser(description='Options to generate '
                                     'a metabolic model')

    if command_class is not None:
        # Command explicitly given, only allow that command
        command_class.init_parser(parser)
        parser.set_defaults(command=command_class)
    else:
        # Discover all available options
        commands = {}
        for entry in pkg_resources.iter_entry_points(
                'psamm.generate_model'):
            canonical = entry.name.lower()
            if canonical not in commands:
                command_class = entry.load()
                commands[canonical] = command_class
            else:
                logger.warning('Option {} was found more than once!'.format(
                    canonical.name))

        # Create parsers for subcommands
        subparsers = parser.add_subparsers(title='Commands',
                                           metavar='command')
        for name, command_class in sorted(iteritems(commands)):
            title, _, _ = command_class.__doc__.partition('\n\n')
            subparser = subparsers.add_parser(
                name, help=title.rstrip('.'),
                formatter_class=argparse.RawDescriptionHelpFormatter,
                description=_trim(command_class.__doc__))
            subparser.set_defaults(command=command_class)
            command_class.init_parser(subparser)

    parsed_args = parser.parse_args(args)
    command = parsed_args.command(parsed_args)
    try:
        command.run()
    except CommandError as e:
        parser.error(text_type(e))


class ReactionEntry(object):
    """Representation of entry in KEGG reaction file"""

    def __init__(self, values, rhea, filemark=None):
        self.values = dict(values)
        if 'entry' not in values:
            raise ParseError2('Missing reaction identifier')
        if 'transport_name' in values:
            self._id = values['transport_name']
        else:
            self._id, _ = values['entry'][0].split(None, 1)
        self._filemark = filemark

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        try:
            return next(self.names)
        except StopIteration:
            return None

    @property
    def names(self):
        if 'name' in self.values:
            for line in self.values['name']:
                for name in line.rstrip(';').split(';'):
                    yield name.strip()

    @property
    def definition(self):
        if 'definition' not in self.values:
            return None
        return self.values['definition'][0]

    @property
    def equation(self):
        if 'equation' not in self.values:
            return None
        return self.values['equation'][0]

    @property
    def enzymes(self):
        if 'enzyme' in self.values:
            for line in self.values['enzyme']:
                for enzyme in line.split():
                    yield enzyme

    @property
    def pathways(self):
        if 'pathway' in self.values:
            for line in self.values['pathway']:
                pathway, name = line.split(None, 1)
                yield pathway, name

    @property
    def comment(self):
        if 'comment' not in self.values:
            return None
        return '\n'.join(self.values['comment'])

    @property
    def tcdb_family(self):
        if 'tcdb_family' not in self.values:
            return None
        return self.values['tcdb_family']

    @property
    def substrates(self):
        if 'substrates' not in self.values:
            return None
        return self.values['substrates']

    @property
    def genes(self):
        if 'genes' not in self.values:
            return None
        return self.values['genes']

    @property
    def orthology(self):
        if 'orthology' in self.values:
            for line in self.values['orthology']:
                split = line.split(None, 1)
                yield split[0]

    def __getitem__(self, name):
        if name not in self.values:
            raise AttributeError('Attribute does not exist: {}'.format(name))
        return self._values[name]

    @property
    def filemark(self):
        return self._filemark

    def __repr__(self):
        return '<ReactionEntry "{}">'.format(self.id)


class CompoundEntry(object):
    """Representation of entry in KEGG compound file"""

    def __init__(self, values, rhea, filemark=None):
        self.values = dict(values)
        if 'entry' not in values:
            raise ParseError2('Missing compound identifier')
        self._id, _ = values['entry'][0].split(None, 1)
        self._filemark = filemark
        self._charge = None
        self._chebi = None
        self._chebi_all = None
        self.rhea = rhea
        self.initialize_charge()

    def initialize_charge(self):
        """
        Sets the _charge, _chebi, and _chebi_all attributes
        'rhea_db' is initialized as a global in generate_model_api
        if --rhea is supplied this funcion looks for rhea_db in the
        global namespace decide if rhea is used
        --- Logic for selecting the best chebi ID ---
        if not using rhea:
            use the first chebi ID given by KEGG
        elif using rhea:
            if all KEGG-chebi IDs map to same ID in rhea:
                use the single ID
            elif KEGG-chebi IDs map to different IDs in rhea:
                use the first chebi ID given by KEGG
            elif the KEGG-chebi IDs don't have mappings in rhea:
                use the first chebi ID given by KEGG
        """
        if self.rhea is not None:
            use_rhea = True
            rhea_db = self.rhea
        else:
            use_rhea = False
        for DB, ID in self.dblinks:
            if DB == "ChEBI":
                id_list = ID.split(" ")
                if use_rhea:
                    rhea_id_list = rhea_db.select_chebi_id(id_list)
                    if len(rhea_id_list) == 0:  # no chebis map to rhea
                        self._chebi = id_list[0]
                        self._chebi_all = id_list
                    elif len(set(rhea_id_list)) == 1:  # chebi map to same rhea
                        self._chebi = rhea_id_list[0]
                        self._chebi_all = set(id_list + [rhea_id_list[0]])
                    else:  # chebis map to different rheas
                        self._chebi = id_list[0]
                        self._chebi_all = set(id_list + rhea_id_list)
                else:  # --rhea not given
                    self._chebi = id_list[0]
                    self._chebi_all = list(id_list)

        # libchebipy update charge and formula
        if self._chebi is not None:
            this_chebi_entity = ChebiEntity(self._chebi)
            try:
                try:
                    # libchebipy sometimes fails with an index error
                    # on the first time running. We have not been able
                    # to fix the source of this error, but catching the
                    # index error and repeating the operation appears to
                    # fix this
                    self._charge = int(this_chebi_entity.get_charge())
                    self.values['formula'] = [this_chebi_entity.get_formula()]
                except IndexError:
                    self._charge = int(this_chebi_entity.get_charge())
                    self.values['formula'] = [this_chebi_entity.get_formula()]
            except ValueError:  # chebi entry has no charge; leave as None
                pass

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        try:
            return next(self.names)
        except StopIteration:
            return None

    @property
    def names(self):
        if 'name' in self.values:
            for line in self.values['name']:
                for name in line.rstrip(';').split(';'):
                    yield name.strip()

    @property
    def reactions(self):
        if 'reaction' in self.values:
            for line in self.values['reaction']:
                for rxnid in line.split():
                    yield rxnid

    @property
    def enzymes(self):
        if 'enzyme' in self.values:
            for line in self.values['enzyme']:
                for enzyme in line.split():
                    yield enzyme

    @property
    def formula(self):
        if 'formula' not in self.values:
            return None
        return self.values['formula'][0]

    @property
    def mol_weight(self):
        if 'mol_weight' not in self.values:
            return None
        return float(self.values['mol_weight'][0])

    @property
    def dblinks(self):
        if 'dblinks' in self.values:
            for line in self.values['dblinks']:
                database, entry = line.split(':', 1)
                yield database.strip(), entry.strip()
        else:
            return None

    @property
    def charge(self):
        return self._charge

    @property
    def chebi(self):
        return self._chebi

    @property
    def chebi_all(self):
        if self._chebi_all is not None:
            return ', '.join(self._chebi_all)
        else:
            return None

    @property
    def comment(self):
        if 'comment' not in self.values:
            return None
        try:
            return '\n'.join(self.values['comment'])
        except TypeError:
            return self.values['comment']

    def __getitem__(self, name):
        if name not in self.values:
            raise AttributeError('Attribute does not exist: {}'.format(name))
        return self._values[name]

    @property
    def filemark(self):
        return self._filemark

    def __repr__(self):
        return '<CompoundEntry "{}">'.format(self.id)


class RheaDb(object):
    """Allows storing and searching Rhea db"""

    def __init__(self, filepath):
        self._values = self._parse_db_from_tsv(filepath)

    def _parse_db_from_tsv(self, filepath):
        db = {}
        with open(filepath, 'r') as f:
            for line in f:
                split = line.split('\t')
                db[split[0]] = split[1]
        return db

    def select_chebi_id(self, id_list):
        rhea_id_list = []
        for x in id_list:
            try:
                rhea_id_list.append(self._values[x])
            except KeyError:
                pass
        return rhea_id_list



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2016  Chao Liu <lcddzyx@gmail.com>
# Copyright 2020 Christopher Powers <c-11060@my.uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Functionality related to gap-filling in general.

This module contains some general functions for preparing models for
gap-filling. Specific gap-filling methods are implemented in the ``gapfill``
and ``fastgapfill`` modules.
"""

from __future__ import unicode_literals

import logging

from six import iteritems

from .metabolicmodel import create_exchange_id, create_transport_id
from .reaction import Reaction, Direction

logger = logging.getLogger(__name__)


def add_all_database_reactions(model, compartments):
    """Add all reactions from database that occur in given compartments.

    Args:
        model: :class:`psamm.metabolicmodel.MetabolicModel`.
    """

    added = set()
    for rxnid in model.database.reactions:
        reaction = model.database.get_reaction(rxnid)
        if all(compound.compartment in compartments
               for compound, _ in reaction.compounds):
            if not model.has_reaction(rxnid):
                added.add(rxnid)
            model.add_reaction(rxnid)

    return added


def add_all_exchange_reactions(model, compartment, allow_duplicates=False):
    """Add all exchange reactions to database and to model.

    Args:
        model: :class:`psamm.metabolicmodel.MetabolicModel`.
    """

    all_reactions = {}
    if not allow_duplicates:
        # TODO: Avoid adding reactions that already exist in the database.
        # This should be integrated in the database.
        for rxnid in model.database.reactions:
            rx = model.database.get_reaction(rxnid)
            all_reactions[rx] = rxnid

    added = set()
    added_compounds = set()
    initial_compounds = set(model.compounds)
    reactions = set(model.database.reactions)
    for model_compound in initial_compounds:
        compound = model_compound.in_compartment(compartment)
        if compound in added_compounds:
            continue

        rxnid_ex = create_exchange_id(reactions, compound)

        reaction_ex = Reaction(Direction.Both, {compound: -1})
        if reaction_ex not in all_reactions:
            model.database.set_reaction(rxnid_ex, reaction_ex)
            reactions.add(rxnid_ex)
        else:
            rxnid_ex = all_reactions[reaction_ex]

        if not model.has_reaction(rxnid_ex):
            added.add(rxnid_ex)
        model.add_reaction(rxnid_ex)
        added_compounds.add(compound)

    return added


def add_all_transport_reactions(model, boundaries, allow_duplicates=False):
    """Add all transport reactions to database and to model.

    Add transport reactions for all boundaries. Boundaries are defined
    by pairs (2-tuples) of compartment IDs. Transport reactions are
    added for all compounds in the model, not just for compounds in the
    two boundary compartments.

    Args:
        model: :class:`psamm.metabolicmodel.MetabolicModel`.
        boundaries: Set of compartment boundary pairs.

    Returns:
        Set of IDs of reactions that were added.
    """

    all_reactions = {}
    if not allow_duplicates:
        # TODO: Avoid adding reactions that already exist in the database.
        # This should be integrated in the database.
        for rxnid in model.database.reactions:
            rx = model.database.get_reaction(rxnid)
            all_reactions[rx] = rxnid

    boundary_pairs = set()
    for source, dest in boundaries:
        if source != dest:
            boundary_pairs.add(tuple(sorted((source, dest))))

    added = set()
    added_pairs = set()
    initial_compounds = set(model.compounds)
    reactions = set(model.database.reactions)
    for compound in initial_compounds:
        for c1, c2 in boundary_pairs:
            compound1 = compound.in_compartment(c1)
            compound2 = compound.in_compartment(c2)
            pair = compound1, compound2
            if pair in added_pairs:
                continue

            rxnid_tp = create_transport_id(reactions, compound1, compound2)

            reaction_tp = Reaction(Direction.Both, {
                compound1: -1,
                compound2: 1
            })
            if reaction_tp not in all_reactions:
                model.database.set_reaction(rxnid_tp, reaction_tp)
                reactions.add(rxnid_tp)
            else:
                rxnid_tp = all_reactions[reaction_tp]

            if not model.has_reaction(rxnid_tp):
                added.add(rxnid_tp)
            model.add_reaction(rxnid_tp)
            added_pairs.add(pair)

    return added


def create_extended_model(model, db_penalty=None, ex_penalty=None,
                          tp_penalty=None, penalties=None):
    """Create an extended model for gap-filling.

    Create a :class:`psamm.metabolicmodel.MetabolicModel` with
    all reactions added (the reaction database in the model is taken
    to be the universal database) and also with artificial exchange
    and transport reactions added. Return the extended
    :class:`psamm.metabolicmodel.MetabolicModel`
    and a weight dictionary for added reactions in that model.

    Args:
        model: :class:`psamm.datasource.native.NativeModel`.
        db_penalty: penalty score for database reactions, default is `None`.
        ex_penalty: penalty score for exchange reactions, default is `None`.
        tb_penalty: penalty score for transport reactions, default is `None`.
        penalties: a dictionary of penalty scores for database reactions.
    """

    # Create metabolic model
    model_extended = model.create_metabolic_model()
    extra_compartment = model.extracellular_compartment

    compartment_ids = set(c.id for c in model.compartments)

    # Add database reactions to extended model
    if len(compartment_ids) > 0:
        logger.info(
            'Using all database reactions in compartments: {}...'.format(
                ', '.join('{}'.format(c) for c in compartment_ids)))
        db_added = add_all_database_reactions(model_extended, compartment_ids)
    else:
        logger.warning(
            'No compartments specified in the model; database reactions will'
            ' not be used! Add compartment specification to model to include'
            ' database reactions for those compartments.')
        db_added = set()

    # Add exchange reactions to extended model
    logger.info(
        'Using artificial exchange reactions for compartment: {}...'.format(
            extra_compartment))
    ex_added = add_all_exchange_reactions(
        model_extended, extra_compartment, allow_duplicates=False)

    # Add transport reactions to extended model
    boundaries = model.compartment_boundaries
    if len(boundaries) > 0:
        logger.info(
            'Using artificial transport reactions for the compartment'
            ' boundaries: {}...'.format(
                '; '.join('{}<->{}'.format(c1, c2) for c1, c2 in boundaries)))
        tp_added = add_all_transport_reactions(
            model_extended, boundaries, allow_duplicates=True)
    else:
        logger.warning(
            'No compartment boundaries specified in the model;'
            ' artificial transport reactions will not be used!')
        tp_added = set()

    # Add penalty weights on reactions
    weights = {}
    if db_penalty is not None:
        weights.update((rxnid, db_penalty) for rxnid in db_added)
    else:
        weights.update((rxnid, 1) for rxnid in db_added)
    if tp_penalty is not None:
        weights.update((rxnid, tp_penalty) for rxnid in tp_added)
    else:
        weights.update((rxnid, 1) for rxnid in tp_added)
    if ex_penalty is not None:
        weights.update((rxnid, ex_penalty) for rxnid in ex_added)
    else:
        weights.update((rxnid, 1) for rxnid in ex_added)

    if penalties is not None:
        for rxnid, penalty in iteritems(penalties):
            weights[rxnid] = penalty
    return model_extended, weights



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""SBML importers."""

import os
import glob
import logging


from ..datasource import sbml
from ..datasource.context import FilePathContext
from ..datasource.entry import (DictCompartmentEntry, DictCompoundEntry,
                                DictReactionEntry)

from ..importer import Importer, ParseError, ModelLoadError

logger = logging.getLogger(__name__)


class BaseImporter(Importer):
    """Base importer for reading metabolic model from an SBML file."""

    def help(self):
        """Print importer help text."""
        print('Source must contain the model definition in SBML format.\n'
              'Expected files in source directory:\n'
              '- *.sbml')

    def _resolve_source(self, source):
        """Resolve source to filepath if it is a directory."""
        if os.path.isdir(source):
            sources = glob.glob(os.path.join(source, '*.sbml'))
            if len(sources) == 0:
                raise ModelLoadError('No .sbml file found in source directory')
            elif len(sources) > 1:
                raise ModelLoadError(
                    'More than one .sbml file found in source directory')
            return sources[0]
        return source

    def _get_reader(self, f):
        raise NotImplementedError('Subclasses must implement _get_reader()')

    def import_model(self, source):
        """Import and return model instance."""
        source = self._resolve_source(source)
        self._context = FilePathContext(source)
        with self._context.open() as f:
            self._reader = self._open_reader(f)

        return self._reader.create_model()


class StrictImporter(BaseImporter):
    """Read metabolic model from an SBML file using strict parser."""

    name = 'SBML-strict'
    title = 'SBML model (strict)'
    generic = True

    def _open_reader(self, f):
        try:
            return sbml.SBMLReader(f, strict=True, ignore_boundary=True)
        except sbml.ParseError as e:
            raise ParseError(e)

    def _translate_compartment(self, entry, new_id):
        return DictCompartmentEntry(entry, new_id)

    def _translate_compound(self, entry, new_id, compartment_map):
        return DictCompoundEntry(entry, new_id)

    def _translate_reaction(
            self, entry, new_id, compartment_map, compound_map):
        return DictReactionEntry(entry, new_id)

    def import_model(self, source):
        """Import and return model instance."""
        model = super(StrictImporter, self).import_model(source)

        # Translate entries into dict-based entries.
        sbml.convert_model_entries(
            model, convert_id=lambda entry: entry.id,
            translate_compartment=self._translate_compartment,
            translate_compound=self._translate_compound,
            translate_reaction=self._translate_reaction)

        return model


class NonstrictImporter(BaseImporter):
    """Read metabolic model from an SBML file using non-strict parser."""

    name = 'SBML'
    title = 'SBML model (non-strict)'
    generic = True

    def _open_reader(self, f):
        try:
            return sbml.SBMLReader(f, strict=False, ignore_boundary=True)
        except sbml.ParseError as e:
            raise ParseError(e)

    def import_model(self, source):
        """Import and return model instance."""
        model = super(NonstrictImporter, self).import_model(source)
        sbml.convert_sbml_model(model)
        return model



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import unittest
from decimal import Decimal

from six import text_type

from psamm.reaction import Reaction, Compound, Direction
from psamm.datasource import reaction


class TestDefaultReactionParser(unittest.TestCase):
    def test_reaction_parse(self):
        r = reaction.parse_reaction(
            '|H2O| + |PPi| => (2) |Phosphate| + (2) |H+|')
        self.assertEqual(r, Reaction(
            Direction.Forward, [(Compound('H2O'), 1), (Compound('PPi'), 1)],
            [(Compound('Phosphate'), 2), (Compound('H+'), 2)]))

    def test_reaction_parse_space(self):
        r = reaction.parse_reaction(
            '(2) |Some compound[c]| <=> '
            'abc[c] + |Some other compound| + (3) def')
        self.assertEqual(r, Reaction(
            Direction.Both, [(Compound('Some compound', 'c'), 2)],
            [(Compound('abc', 'c'), 1),
             (Compound('Some other compound'), 1), (Compound('def'), 3)]))

    def test_reaction_parse_with_decimal(self):
        r = reaction.parse_reaction('|H2| + (0.5) |O2| => |H2O|')
        self.assertEqual(r, Reaction(
            Direction.Forward,
            [(Compound('H2'), 1), (Compound('O2'), Decimal('0.5'))],
            [(Compound('H2O'), 1)]))

    def test_reaction_parse_with_compartment(self):
        r = reaction.parse_reaction('(2) |H2| + |O2| => (2) |H2O[e]|')
        self.assertEqual(r, Reaction(
            Direction.Forward, [(Compound('H2'), 2), (Compound('O2'), 1)],
            [(Compound('H2O', compartment='e'), 2)]))

    def test_reaction_parse_with_multichar_compartment(self):
        r = reaction.parse_reaction(
            '(2) |H2[C_c]| + |O2[C_c]| => (2) |H2O[C_e]|')
        self.assertEqual(r, Reaction(
            Direction.Forward,
            [(Compound('H2', compartment='C_c'), 2),
             (Compound('O2', compartment='C_c'), 1)],
            [(Compound('H2O', compartment='C_e'), 2)]))

    def test_reaction_parse_raw_compound_id(self):
        r = reaction.parse_reaction('(2) cpd00001 => cpd00002')
        self.assertEqual(r, Reaction(
            Direction.Forward, [(Compound('cpd00001'), 2)],
            [(Compound('cpd00002'), 1)]))

    def test_reaction_parse_raw_compound_id_with_typo(self):
        r = reaction.parse_reaction('(2) cpd00001 => cdp00002')
        self.assertEqual(r, Reaction(
            Direction.Forward, [(Compound('cpd00001'), 2)],
            [(Compound('cdp00002'), 1)]))

    def test_reaction_parse_raw_compound_id_in_compartment(self):
        r = reaction.parse_reaction('(2) cpd00001 => cpd00002[e]')
        self.assertEqual(r, Reaction(
            Direction.Forward, [(Compound('cpd00001'), 2)],
            [(Compound('cpd00002', 'e'), 1)]))

    def test_reaction_parse_bidirectional(self):
        r = reaction.parse_reaction('(1) |cpd1| <=> (2) cpd2 + |cpd3|')
        self.assertEqual(r, Reaction(
            Direction.Both, [(Compound('cpd1'), 1)],
            [(Compound('cpd2'), 2), (Compound('cpd3'), 1)]))

    def test_reaction_str(self):
        r = Reaction(Direction.Reverse, [(Compound('H2O'), 2)],
                     [(Compound('H2'), 2), (Compound('O2'), 1)])
        self.assertEqual(text_type(r), '(2) H2O <= (2) H2 + O2')


class TestSudenSimple(unittest.TestCase):
    def setUp(self):
        arrows = (
            ('<=>', Direction.Both),
            ('-->', Direction.Forward)
        )
        self.parser = reaction.ReactionParser(arrows=arrows)

    def test_sudensimple_parse(self):
        r = self.parser.parse('1 H2O + 1 PPi <=> 2 Phosphate + 2 proton')
        self.assertEqual(r, Reaction(
            Direction.Both, [(Compound('H2O'), 1), (Compound('PPi'), 1)],
            [(Compound('Phosphate'), 2), (Compound('proton'), 2)]))

    def test_sudensimple_parse_with_implicit_count(self):
        r = self.parser.parse('H2O + PPi <=> 2 Phosphate + 2 proton')
        self.assertEqual(r, Reaction(
            Direction.Both, [(Compound('H2O'), 1), (Compound('PPi'), 1)],
            [(Compound('Phosphate'), 2), (Compound('proton'), 2)]))

    def test_sudensimple_parse_with_decimal(self):
        r = self.parser.parse('1 H2 + 0.5 O2 <=> 1 H2O')
        self.assertEqual(r, Reaction(
            Direction.Both, [(Compound('H2'), 1),
                             (Compound('O2'), Decimal('0.5'))],
            [(Compound('H2O'), 1)]))


class TestMetNet(unittest.TestCase):
    def setUp(self):
        arrows = (
            ('<==>', Direction.Both),
            ('-->', Direction.Forward)
        )
        self.parser = reaction.ReactionParser(arrows=arrows, parse_global=True)

    def test_metnet_parse_with_global_compartment(self):
        r = self.parser.parse('[c] : akg + ala-L <==> glu-L + pyr')
        self.assertEqual(r, Reaction(
            Direction.Both,
            [(Compound('akg', 'c'), 1), (Compound('ala-L', 'c'), 1)],
            [(Compound('glu-L', 'c'), 1), (Compound('pyr', 'c'), 1)]))

    def test_metnet_parse_with_local_compartment(self):
        r = self.parser.parse(
            '(2) ficytcc553[c] + so3[c] + h2o[c] -->'
            ' (2) focytcc553[c] + so4[c] + (2) h[e]')
        self.assertEqual(r, Reaction(
            Direction.Forward,
            [(Compound('ficytcc553', 'c'), 2), (Compound('so3', 'c'), 1),
             (Compound('h2o', 'c'), 1)],
            [(Compound('focytcc553', 'c'), 2), (Compound('so4', 'c'), 1),
             (Compound('h', 'e'), 2)]))

    def test_metnet_parse_global_with_colon_in_name(self):
        r = self.parser.parse(
            '[c] : fdxr-4:2 + h + nadp <==> fdxo-4:2 + nadph')
        self.assertEqual(r, Reaction(
            Direction.Both,
            [(Compound('fdxr-4:2', 'c'), 1), (Compound('h', 'c'), 1),
             (Compound('nadp', 'c'), 1)],
            [(Compound('fdxo-4:2', 'c'), 1), (Compound('nadph', 'c'), 1)]))

    def test_metnet_parse_local_with_colon_in_name(self):
        r = self.parser.parse(
            'fdxr-4:2[c] + h[c] + nadp[c] <==> fdxo-4:2[c] + nadph[c]')
        self.assertEqual(r, Reaction(
            Direction.Both,
            [(Compound('fdxr-4:2', 'c'), 1), (Compound('h', 'c'), 1),
             (Compound('nadp', 'c'), 1)],
            [(Compound('fdxo-4:2', 'c'), 1), (Compound('nadph', 'c'), 1)]))

    def test_metnet_parse_with_numeric_id(self):
        r = self.parser.parse('[c] : 3pg + atp <==> 13dpg + adp')
        self.assertEqual(r, Reaction(
            Direction.Both,
            [(Compound('3pg', 'c'), 1), (Compound('atp', 'c'), 1)],
            [(Compound('13dpg', 'c'), 1), (Compound('adp', 'c'), 1)]))


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2021 Christopher Powers <c-11060@uri.edu>
# Copyright 2022 Jason Vailionis <jason_vailionis@uri.edu>

import unittest
import sys
import os
from psamm import mapmaker
from psamm import generate_model
from psamm import generate_biomass
from pkg_resources import resource_filename
from psamm.datasource.native import NativeModel, ModelReader
from psamm.datasource.reaction import Reaction, Compound, Direction
from collections import defaultdict
import re
import yaml
import pandas as pd
import numpy as np
from io import StringIO


if sys.version_info.minor < 6:
    raise unittest.SkipTest("not compatible with py3.5")

class TestGenerateTransporters(unittest.TestCase):

    def test_parse_orthology(self):
        test_egg = "gene1\ta\tb\tc\td\te\tf\tg\th\ti\t2.3.3.1\tK01647\tk"
        test_egg += "\tl\tR00351\tm\tn\t2.A.49.5.2\tp\tq\n"
        outfile = open("testin.tsv", "w")
        outfile.write(test_egg)
        outfile.close()
        asso = generate_model.parse_orthology("testin.tsv", "tcdb", None)
        self.assertTrue(len(asso) == 1)
        self.assertTrue(asso["2.A.49.5.2"] == ["gene1"])

    def test_transporters(self):
        asso = {"R00351": ["gene1"], "R00375": ["gene2"]}
        generate_model.create_model_api(".", asso, False, False, "c")
        substrate = resource_filename("psamm",
                                      "external-data/tcdb_substrates.tsv")
        family = resource_filename("psamm",
                                   "external-data/tcdb_families.tsv")

        # read in the model and build a dictionary of Chebi IDs
        mr = ModelReader.reader_from_path(".""")
        nm = mr.create_model()
        mm = nm.create_metabolic_model()
        chebi_dict = defaultdict(lambda: [])
        for cpd in nm.compounds:
            if "ChEBI" in cpd.__dict__["_properties"]:
                chebi = re.split(" ", cpd.__dict__["_properties"]["ChEBI"])
                for i in chebi:
                    chebi_dict["CHEBI:{}".format(i)].append(cpd.id)
        # Read in the reaction substrates
        tp_sub_dict = defaultdict(lambda: [])
        with open(substrate, "r") as infile:
            for line in infile:
                line = line.rstrip()
                listall = re.split("\t", line)
                substrates = re.split(r"\|", listall[1])
                sub_out = []
                for i in substrates:
                    sub_out.append(re.split(";", i)[0])
                tp_sub_dict[listall[0]] = sub_out

        # read in the reaction families
        tp_fam_dict = defaultdict(lambda: "")
        with open(family, "r", encoding="utf8") as infile:
            for line in infile:
                line = line.rstrip()
                listall = re.split("\t", line)
                tp_fam_dict[listall[0]] = listall[1]
        tp_asso = {"2.A.11.1.6": ["gene2"], "1.A.8.9.20": ["gene3"]}
        generate_model.gen_transporters(".", tp_asso, tp_sub_dict,
                                        tp_fam_dict, chebi_dict, "c", "e")
        rxn = open("transporters.yaml", "r").read()
        self.assertIn("TP_2.A.11.1.6", rxn)
        self.assertIn("[c] <=> [e]", rxn)
        self.assertIn("CHEBI:34982", rxn)
        self.assertIn("CHEBI:3308", rxn)
        self.assertIn("CHEBI:50744", rxn)
        self.assertIn("C00001_diff", rxn)
        self.assertIn("[c] <=> [e]", rxn)
        self.assertIn("The Major Intrinsic Protein (MIP) Family", rxn)
        self.assertIn("CHEBI:15377", rxn)
        self.assertIn("C00001[c] <=> C00001[e]", rxn)
        # check different compartments
        generate_model.gen_transporters(".", tp_asso, tp_sub_dict,
                                        tp_fam_dict, chebi_dict, "d", "f")
        rxn = open("transporters.yaml", "r").read()
        self.assertIn("C00001[d] <=> C00001[f]", rxn)
        self.assertIn("[d] <=> [f]", rxn)
        os.remove("log.tsv")
        os.remove("reactions.yaml")
        os.remove("reactions_generic.yaml")
        os.remove("model.yaml")
        os.remove("compounds.yaml")
        os.remove("compounds_generic.yaml")
        os.remove("gene-association.tsv")
        os.remove("gene-association_generic.tsv")
        os.remove("model_def.tsv")
        os.remove("transporters.yaml")
        os.remove("transporter_log.tsv")


class TestGenerateDatabase(unittest.TestCase):
    """
    Test cases largely designed to test download of information
    from KEGG through biopython and generate a reaction database
    properly. Note that if these start to fail, one of the avenues
    to check for errors is the composition of the ec, ko, reaction, or
    compound representation in KEGG.
    """


    def test_overall(self):
        # Check ability to create expected output files
        asso = {"R00351": ["gene1"], "R00375": ["gene2"]}
        generate_model.create_model_api(".", asso, False, False, "c")
        self.assertTrue(os.path.exists("log.tsv"))
        self.assertTrue(os.path.exists("reactions.yaml"))
        self.assertTrue(os.path.exists("compounds.yaml"))
        self.assertTrue(os.path.exists("model.yaml"))
        self.assertTrue(os.path.exists("gene-association.tsv"))
        self.assertTrue(os.path.exists("gene-association_generic.tsv"))
        self.assertTrue(os.path.exists("reactions_generic.yaml"))
        self.assertTrue(os.path.exists("compounds_generic.yaml"))
        # Check contents of the created files
        mr = ModelReader.reader_from_path(".")
        nm = mr.create_model()
        mm = nm.create_metabolic_model()
        self.assertIn("R00351", set(mm.reactions))
        self.assertNotIn("R00375", set(mm.reactions))
        self.assertIn(Compound("C00010", "c"), set(mm.compounds))
        self.assertNotIn(Compound("C00039", "c"), set(mm.compounds))
        self.assertIn("2.3.3.1", nm.reactions["R00351"].
                      __dict__["_properties"]["enzymes"])
        self.assertIn("Glyoxylate and dicarboxylate metabolism",
                      nm.reactions["R00351"].
                      __dict__["_properties"]["pathways"])
        self.assertIn("K01647", nm.reactions["R00351"].
                      __dict__["_properties"]["orthology"])
        self.assertEqual("C6H5O7", nm.compounds["C00158"].
                         __dict__["_properties"]["formula"])
        self.assertEqual("16947", nm.compounds["C00158"].
                         __dict__["_properties"]["ChEBI"])
        self.assertEqual(-3, nm.compounds["C00158"].
                         __dict__["_properties"]["charge"])
        # Check for relevant entries in the generic and other files
        cpd = open("compounds_generic.yaml", "r").read()
        self.assertIn("C00039", cpd)
        rxn = open("reactions_generic.yaml", "r").read()
        self.assertIn("R00375", rxn)
        ga = open("gene-association.tsv", "r").read()
        self.assertIn("R00351", ga)
        ga = open("gene-association_generic.tsv", "r").read()
        self.assertIn("R00375", ga)
        os.remove("log.tsv")
        os.remove("reactions.yaml")
        os.remove("reactions_generic.yaml")
        os.remove("model.yaml")
        os.remove("compounds.yaml")
        os.remove("compounds_generic.yaml")
        os.remove("gene-association.tsv")
        os.remove("gene-association_generic.tsv")
        os.remove("model_def.tsv")

    def test_Rhea(self):
        # Test that the rhea flag properly captures charge of atp
        rhea_db = generate_model.RheaDb(resource_filename("psamm",
                                                          "external-data/"
                                                          "chebi_pH7_3_"
                                                          "mapping.tsv"))
        cpd = ["C00002"]
        cpd_out = generate_model._download_kegg_entries(".", cpd, None,
                                                        generate_model.
                                                        CompoundEntry,
                                                        context=None)
        cpd_out = list(cpd_out)
        nonRhea = list(generate_model.model_compounds(cpd_out))
        self.assertTrue(nonRhea[0]["charge"] == 0)
        cpd_out = generate_model._download_kegg_entries(".", cpd, rhea_db,
                                                        generate_model.
                                                        CompoundEntry,
                                                        context=None)
        cpd_out = list(cpd_out)
        Rhea = list(generate_model.model_compounds(cpd_out))
        self.assertTrue(Rhea[0]["charge"] == -4)
        os.remove("log.tsv")

    def test_EC_download(self):
        # Test when EC has one reaction
        rxn_mapping = {"2.3.3.1": ["Gene1"]}
        ec = generate_model.parse_rxns_from_EC(rxn_mapping, ".", False)
        self.assertTrue(len(ec) == 1)
        self.assertTrue(len(ec["R00351"]) == 1)
        self.assertTrue("R00351" in ec)
        self.assertTrue(ec["R00351"] == ["Gene1"])
        # Test when EC has multiple reactions
        rxn_mapping = {"4.2.1.3": ["Gene1"]}
        ec = generate_model.parse_rxns_from_EC(rxn_mapping, ".", False)
        self.assertTrue(len(ec) == 3)
        self.assertTrue("R01324" in ec)
        self.assertTrue("R01325" in ec)
        self.assertTrue("R01900" in ec)
        self.assertTrue(ec["R01324"] == ["Gene1"])
        self.assertTrue(ec["R01325"] == ["Gene1"])
        self.assertTrue(ec["R01900"] == ["Gene1"])
        # Test for multiple genes
        rxn_mapping = {"2.3.3.1": ["Gene1", "Gene2"]}
        ec = generate_model.parse_rxns_from_EC(rxn_mapping, ".", False)
        self.assertTrue(len(ec) == 1)
        self.assertTrue(len(ec["R00351"]) == 2)
        self.assertTrue("R00351" in ec)
        self.assertTrue(ec["R00351"] == ["Gene1", "Gene2"])
        os.remove("log.tsv")

    def test_KO_download(self):
        # Test when EC has one reaction
        rxn_mapping = {"K01647": ["Gene1"]}
        ko = generate_model.parse_rxns_from_KO(rxn_mapping, ".", False)
        self.assertTrue(len(ko) == 1)
        self.assertTrue(len(ko["R00351"]) == 1)
        self.assertTrue("R00351" in ko)
        self.assertTrue(ko["R00351"] == ["Gene1"])
        # Test when EC has multiple reactions
        rxn_mapping = {"K01681": ["Gene1"]}
        ko = generate_model.parse_rxns_from_KO(rxn_mapping, ".", False)
        self.assertTrue(len(ko) == 3)
        self.assertTrue("R01324" in ko)
        self.assertTrue("R01325" in ko)
        self.assertTrue("R01900" in ko)
        self.assertTrue(ko["R01324"] == ["Gene1"])
        self.assertTrue(ko["R01325"] == ["Gene1"])
        self.assertTrue(ko["R01900"] == ["Gene1"])
        # Test for multiple genes
        rxn_mapping = {"K01647": ["Gene1", "Gene2"]}
        ko = generate_model.parse_rxns_from_KO(rxn_mapping, ".", False)
        self.assertTrue(len(ko) == 1)
        self.assertTrue(len(ko["R00351"]) == 2)
        self.assertTrue("R00351" in ko)
        self.assertTrue(ko["R00351"] == ["Gene1", "Gene2"])
        os.remove("log.tsv")

    def test_model_compounds(self):
        # Tests that compounds are properly sorted into generic compounds
        # with the proper attributes.
        cpd = ["C02987", "C00001"]
        cpd_out = generate_model._download_kegg_entries(".", cpd, None,
                                                        generate_model.
                                                        CompoundEntry,
                                                        context=None)
        cpd_out = list(cpd_out)
        generic = generate_model.check_generic_compounds(cpd_out)
        generic_out = generate_model._download_kegg_entries(".", generic, None,
                                                            generate_model.
                                                            CompoundEntry,
                                                            context=None)
        generic_out = generate_model.model_generic_compounds(list(generic_out))
        generic_out = list(generic_out)
        self.assertTrue(len(cpd_out) == 2)
        self.assertTrue(len(generic_out) == 1)
        self.assertTrue("C02987" == cpd_out[0].id)
        self.assertTrue("C00001" == cpd_out[1].id)
        self.assertTrue("C02987" == generic_out[0]["id"])
        self.assertTrue("L-Glutamyl-tRNA(Glu)" == generic_out[0]["name"])
        self.assertTrue("C20H28N6O13PR(C5H8O6PR)n" ==
                        generic_out[0]["formula"])
        self.assertTrue("29157" == generic_out[0]["ChEBI"])
        cpd_out = list(generate_model.model_compounds(cpd_out))
        self.assertTrue(len(cpd_out) == 1)
        self.assertTrue("C00001" == cpd_out[0]["id"])
        self.assertTrue("H2O" == cpd_out[0]["name"])
        self.assertTrue("H2O" == cpd_out[0]["formula"])
        self.assertTrue("15377" == cpd_out[0]["ChEBI"])

    def test_generic_compoundID(self):
        # Test that the download of compounds works
        cpd = ["C02987", "C00001"]
        cpd_out = generate_model._download_kegg_entries(".", cpd, None,
                                                        generate_model.
                                                        CompoundEntry,
                                                        context=None)
        cpd_out = generate_model.check_generic_compounds(list(cpd_out))
        cpd_out = list(cpd_out)
        self.assertTrue(len(cpd_out) == 1)
        self.assertTrue("C02987" in cpd_out)
        self.assertTrue("C00001" not in cpd_out)
        os.remove("log.tsv")

    def test_Compound_Download(self):
        # Test that the download of compounds works
        cpd = ["C00001"]
        cpd_out = generate_model._download_kegg_entries(".", cpd, None,
                                                        generate_model.
                                                        CompoundEntry,
                                                        context=None)
        cpd_out = list(cpd_out)
        self.assertTrue(len(cpd_out) == 1)
        os.remove("log.tsv")

    def test_rxn_clean(self):
        # Test teh function that reformats stoichiometry
        rxn = ["R04347"]
        rxn_out = generate_model._download_kegg_entries(".", rxn, None,
                                                        generate_model.
                                                        ReactionEntry,
                                                        context=None)
        rxn_out, gen= generate_model.clean_reaction_equations(list(rxn_out))
        self.assertTrue("R04347" in gen)

    def test_Compound_Contents(self):
        # Test that the downloaded compound contains the relevant information
        cpd = ["C00001"]
        cpd_out = generate_model._download_kegg_entries(".", cpd, None,
                                                        generate_model.
                                                        CompoundEntry,
                                                        context=None)
        cpd_out = list(cpd_out)
        self.assertTrue(len(cpd_out) == 1)
        self.assertEqual(list(cpd_out)[0].id, "C00001")
        self.assertEqual(list(cpd_out)[0].name, "H2O")
        self.assertEqual(list(cpd_out)[0].formula, "H2O")
        self.assertEqual(list(cpd_out)[0].mol_weight, 18.0153)
        self.assertEqual(list(cpd_out)[0].chebi, "15377")
        self.assertEqual(list(cpd_out)[0].charge, 0)
        os.remove("log.tsv")

    def test_Reaction_Download(self):
        # Test that the download of reactions works
        rxn = ["R00351"]
        rxn_out = generate_model._download_kegg_entries(".", rxn, None,
                                                        generate_model.
                                                        ReactionEntry,
                                                        context=None)
        self.assertTrue(len(list(rxn_out)) == 1)
        os.remove("log.tsv")

    def test_Reaction_Contents(self):
        # Test that the downloaded reaction contains the relevant information
        rxn = ["R00351"]
        rxn_out = generate_model._download_kegg_entries(".", rxn, None,
                                                        generate_model.
                                                        ReactionEntry,
                                                        context=None)
        rxn_out = list(rxn_out)
        self.assertEqual(rxn_out[0].id, "R00351")
        self.assertEqual(rxn_out[0].name, "acetyl-CoA:oxaloacetate "
                         "C-acetyltransferase (thioester-hydrolysing)")
        self.assertEqual(rxn_out[0].definition, "Citrate + CoA <=> Acetyl-CoA"
                         " + H2O + Oxaloacetate")
        self.assertEqual(rxn_out[0].equation, "C00158 + C00010 <=> C00024 "
                         "+ C00001 + C00036")
        self.assertEqual(list(rxn_out[0].enzymes),
                         ["2.3.3.1", "2.3.3.3", "2.3.3.16"])
        self.assertIsNone(rxn_out[0].tcdb_family)
        self.assertIsNone(rxn_out[0].substrates)
        self.assertIsNone(rxn_out[0].genes)
        self.assertTrue(len(list(rxn_out[0].pathways)) == 8)
        self.assertEqual(list(rxn_out[0].pathways)[0],
                         ("rn00020", "Citrate cycle (TCA cycle)"))
        os.remove("log.tsv")

    def test_Model_Reactions(self):
        # Tests the conversion of the reactions object to a dictionary format
        rxn = ["R00351"]
        rxn_out = []
        for entry in generate_model._download_kegg_entries(".", rxn, None,
                                                           generate_model.
                                                           ReactionEntry,
                                                           context=None):
            rxn_out.append(entry)
        rxn_model = generate_model.model_reactions(rxn_out)
        rxn_model = list(rxn_model)
        self.assertEqual(rxn_out[0].id, rxn_model[0]["id"])
        self.assertEqual(rxn_out[0].name, rxn_model[0]["name"])
        self.assertEqual(rxn_out[0].definition, rxn_model[0]
                         ["KEGG_definition"])
        self.assertEqual(rxn_out[0].equation, rxn_model[0]["equation"])
        self.assertEqual(list(rxn_out[0].enzymes), rxn_model[0]["enzymes"])
        path = []
        for i in list(rxn_out[0].pathways):
            path.append(i[1])
        self.assertEqual(len(list(rxn_out[0].pathways)), len(rxn_model[0]
                         ["pathways"]))
        self.assertEqual(path, rxn_model[0]["pathways"])
        os.remove("log.tsv")

    def test_parseOrtho(self):
        # Test the ability to parse the defaul eggnog output
        test_egg = "gene1\ta\tb\tc\td\te\tf\tg\th\ti\t2.3.3.1\t"
        test_egg += "K01647\tk\tl\tR00351\tm\tn\to\tp\tq\n"
        outfile = open("testin.tsv", "w")
        outfile.write(test_egg)
        outfile.close()
        asso = generate_model.parse_orthology("testin.tsv", "R", None)
        self.assertTrue(len(asso) == 1)
        self.assertTrue(asso["R00351"] == ["gene1"])
        asso = generate_model.parse_orthology("testin.tsv", "KO", None)
        self.assertTrue(len(asso) == 1)
        self.assertTrue(asso["K01647"] == ["gene1"])
        asso = generate_model.parse_orthology("testin.tsv", "EC", None)
        self.assertTrue(len(asso) == 1)
        self.assertTrue(asso["2.3.3.1"] == ["gene1"])
        asso = generate_model.parse_orthology("testin.tsv", "R", 2)
        self.assertTrue(len(asso) == 1)
        self.assertTrue(asso["a"] == ["gene1"])
        os.remove("testin.tsv")


class TestGenerateBiomass(unittest.TestCase):

    def test_cpd_database_load(self):
        df = generate_biomass.load_compound_data()
        self.assertTrue(df.shape == (79, 7))
        config = generate_biomass.make_template(df)
        self.assertTrue(config.shape == (79, 3))
        config.custom_id = list(range(79))  # config w/ custom ids from 0 to 78
        config.to_csv("config_file.csv", index=False)
        modified_df = generate_biomass.apply_config(df, "config_file.csv")
        self.assertTrue(modified_df.shape == (79, 8))
        self.assertTrue(modified_df.loc["biomass", "id"] == 0)
        self.assertTrue(modified_df.loc["C02554", "id"] == 78)
        os.remove("config_file.csv")

    def test_model_parsing(self):
        # THIS will include the way to build a fake model
        # Then, test the load_model, check_missing_cpds, fix_cell_compartment,
        # and update_model_yaml methods
        yaml_args = {"default_flow_style": False,
                     "sort_keys": False,
                     "encoding": "utf-8",
                     "allow_unicode": True,
                     "width": float("inf")}
        test_model_dict = {"default_compartment": None, "biomass": None,
                           "compounds": [{"include": "compounds.yaml"}],
                           "reactions": [{"include": "reactions.yaml"}],
                           "model": [{"include": "model_def.tsv"}]}
        tmp_model_path = "tmp_model.yaml"
        df = generate_biomass.load_compound_data()
        with open(tmp_model_path, "w") as f:
            yaml.dump(test_model_dict, f, **yaml_args)
        model_dict = generate_biomass.load_model(tmp_model_path)
        self.assertTrue(test_model_dict == model_dict)
        # FAKE compounds.yaml which contains one of the 79 biomass compounds
        # Testing if generate_biomass.yaml has the other 78 compounds
        with open("compounds.yaml", "w") as f:
            f.write("- id: C02554")
        generate_biomass.check_missing_cpds(
            tmp_model_path, df, None, yaml_args)
        # Read in the biomass_compounds.yaml and make sure theres 78 cpds
        # matching the first 78 in the df
        with open("biomass_compounds.yaml", "r") as f:
            matches = re.findall("- id:.*\n", f.read())
            cpds = [x.rstrip("\n").replace("- id: ", "") for x in matches]
        self.assertTrue(len(cpds) == 78)
        self.assertTrue(list(df.id)[:-1] == cpds)
        # Check fix_cell_compartment()
        model_dict["default_compartment"] = None
        self.assertTrue(
            generate_biomass.fix_cell_compartment(model_dict) == "c")
        model_dict["default_compartment"] = "x"
        self.assertTrue(
            generate_biomass.fix_cell_compartment(model_dict) == "x")
        del model_dict["default_compartment"]
        self.assertTrue(
            generate_biomass.fix_cell_compartment(model_dict) == "c")
        # Check update_model_yaml()
        test_model_dict["compounds"].append(
            {"include": "./biomass_compounds.yaml"})
        test_model_dict["reactions"].append(
            {"include": "./biomass_reactions.yaml"})
        test_model_dict["biomass"] = "biomass_name"
        test_model_dict["default_compartment"] = "c"
        generate_biomass.update_model_yaml(
            tmp_model_path, model_dict, "c", "biomass_name", yaml_args)
        updated_model_dict = generate_biomass.load_model(tmp_model_path)
        self.assertTrue(test_model_dict == updated_model_dict)
        os.remove("biomass_compounds.yaml")
        os.remove(tmp_model_path)
        os.remove("compounds.yaml")

    def test_calc_stoichiometry(self):
        df = pd.DataFrame({"mol_weight": [1, 2, 3, 3]})
        out = generate_biomass.calc_stoichiometry_df(df, [2, 2, 3, 3])
        self.assertTrue(out.shape == (4, 5))
        self.assertTrue(list(out.columns) == ["mol_weight", "counts",
                                              "composition", "mass",
                                              "stoichiometry"])
        np.testing.assert_almost_equal(out.stoichiometry,
                                       [0.0833333, 0.0833333, 0.125, 0.125])

    def test_dna_entry(self):
        from Bio import SeqIO
        with StringIO(">seq1\nAAGGGT\n>seq2\nAGCT") as seqfile:
            genome = {seq.name: seq.upper()
                      for seq in SeqIO.parse(seqfile, "fasta")}
        df = generate_biomass.load_compound_data()
        dna_output = generate_biomass.count_DNA(genome, df, "c")
        print(dna_output)
        proper_output = {"id": "dna_met", "name": "DNA",
                         "equation": "(0.613065) C00131[c] + (0.40871) "
                         "C00459[c] + (0.81742) C00286[c] + (0.204355) "
                         "C00458[c] + (4.0871) C00002[c] + (4.0871) C00001[c] "
                         "=> dna[c] + (4.0871) C00008[c] + (4.0871) C00009[c] "
                         "+ (2.04355) C00013[c]", "pathways": ["Biomass"]}
        self.assertTrue(dna_output == proper_output)

    def test_rna_entry(self):
        from Bio import SeqIO
        df = generate_biomass.load_compound_data()
        with StringIO(">seq1\nAAGGGT\n>seq2\nAGCT") as fna:
            genome = {seq.name: seq.upper()
                      for seq in SeqIO.parse(fna, "fasta")}
        with open("temp.gff", "w") as gff:
            gff.write("seq1\tProtein Homology\tCDS\t2\t6\t.\t+\t0\ttest\n" +
                      "seq2\tProtein Homology\tCDS\t2\t4\t.\t-\t0\ttest")
        rna_output = generate_biomass.count_RNA(genome, "temp.gff", df, "c")
        proper_output = {"id": "rna_met", "name": "RNA",
                         "equation": "(4.248569) C00002[c] + (0.326813) "
                         "C00075[c] + (1.307252) C00044[c] + (3.921756) "
                         "C00001[c] => rna[c] + (3.921756) C00008[c] + "
                         "(3.921756) C00009[c] + (1.960878) C00013[c]",
                         "pathways": ["Biomass"]}
        self.assertTrue(rna_output == proper_output)
        os.remove("temp.gff")

    def test_prot_entry(self):
        from Bio import SeqIO
        df = generate_biomass.load_compound_data()
        with StringIO(">seq1\nLEW\n>seq2\nMIG") as faa:
            proteome = {seq.name: seq.upper()
                        for seq in SeqIO.parse(faa, "fasta")}
        prot_output = generate_biomass.count_Prot(proteome, df, "c")
        proper_output = {"id": "pro_met", "name": "Protein",
                         "equation": "(1.194786) C02987[c] + (1.194786) "
                         "C02412[c] + (1.194786) C03127[c] + (1.194786) "
                         "C02047[c] + (1.194786) C02430[c] + (1.194786) "
                         "C03512[c] + (14.337431) C00044[c] + (14.337431) "
                         "C00001[c] => (1.194786) C01641[c] + (1.194786) "
                         "C01642[c] + (1.194786) C01644[c] + (1.194786) "
                         "C01645[c] + (1.194786) C01647[c] + (1.194786) "
                         "C01652[c] + protein[c] + (14.337431) C00035[c] + "
                         "(14.337431) C00009[c]", "pathways": ["Biomass"]}
        self.assertTrue(prot_output == proper_output)

    def test_bio_entry(self):
        df = generate_biomass.load_compound_data()
        bio_output = generate_biomass.return_biomass_rxn(df, "biomass", "c")
        proper_output = {"id": "biomass", "name": "Biomass",
                         "equation": "dna[c] + rna[c] + protein[c] => "
                         "biomass[c]", "pathways": ["Biomass"]}
        self.assertTrue(bio_output == proper_output)

    def test_sink_entry(self):
        df = generate_biomass.load_compound_data()
        sink_output = generate_biomass.return_bio_sink_rxn(df, "c")
        proper_output = {"id": "sink_biomass", "name": "Biomass accumulation",
                         "equation": "biomass[c] =>", "pathways": ["Biomass"]}
        self.assertTrue(sink_output == proper_output)

    def test_tRNA_entries(self):
        df = generate_biomass.load_compound_data()
        trna_output = generate_biomass.return_trna_rxns(df, "c")
        self.assertTrue(len(trna_output) == 20)
        proper_output_lastaa = {"id": "R03665",
                                "name": "L-Valine:tRNAVal ligase "
                                "(AMP-forming)",
                                "equation": "C00002[c] + C00183[c] + C01653[c]"
                                " <=> C00020[c] + C00013[c] + C02554[c]",
                                "enzyme": "['6.1.1.9']",
                                "pathways": ["Biomass"]}
        self.assertTrue(trna_output[19] == proper_output_lastaa)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import logging
from six import iteritems

from ..command import Command, MetabolicMixin, SolverCommandMixin
from ..gapfill import gapfind, GapFillError
from ..lpsolver import lp

logger = logging.getLogger(__name__)


class GapCheckCommand(MetabolicMixin, SolverCommandMixin, Command):
    """Check for compound production gaps in model."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--method', choices=['gapfind', 'prodcheck', 'sinkcheck'],
            default='prodcheck',
            help='Method to use for gap checking (default: prodcheck)')
        parser.add_argument(
            '--no-implicit-sinks', action='store_true',
            help='Do not include implicit sinks when gap-filling')
        parser.add_argument(
            '--unrestricted-exchange', action='store_true',
            help='Remove all limits on exchange reactions while gap checking')
        parser.add_argument(
            '--exclude-extracellular', action='store_true',
            help='Exclude extracellular compounds from the result')
        parser.add_argument(
            '--epsilon', type=float, default=1e-5,
            help='Threshold for compound production')

        super(GapCheckCommand, cls).init_parser(parser)

    def run(self):
        # Load compound information
        def compound_name(id):
            if id not in self._model.compounds:
                return id
            return self._model.compounds[id].properties.get('name', id)

        extracellular_comp = self._model.extracellular_compartment
        epsilon = self._args.epsilon
        v_max = float(self._model.default_flux_limit)

        implicit_sinks = not self._args.no_implicit_sinks

        if self._args.unrestricted_exchange:
            # Allow all exchange reactions with no flux limits
            for reaction in self._mm.reactions:
                if self._mm.is_exchange(reaction):
                    del self._mm.limits[reaction].bounds

        logger.info('Searching for blocked compounds...')

        if self._args.method == 'gapfind':
            # Run GapFind on model
            solver = self._get_solver(integer=True)
            try:
                blocked = sorted(gapfind(
                    self._mm, solver=solver, epsilon=epsilon, v_max=v_max,
                    implicit_sinks=implicit_sinks))
            except GapFillError as e:
                self._log_epsilon_and_fail(epsilon, e)
        elif self._args.method == 'prodcheck':
            # Run production check on model
            solver = self._get_solver()
            blocked = self.run_reaction_production_check(
                self._mm, solver=solver, threshold=self._args.epsilon,
                implicit_sinks=implicit_sinks)
        elif self._args.method == 'sinkcheck':
            # Run sink check on model
            solver = self._get_solver()
            blocked = self.run_sink_check(
                self._mm, solver=solver, threshold=self._args.epsilon,
                implicit_sinks=implicit_sinks)
        else:
            self.argument_error('Invalid method: {}'.format(self._args.method))

        # Show result
        count = 0
        for compound in blocked:
            if (self._args.exclude_extracellular and
                    compound.compartment == extracellular_comp):
                continue

            count += 1
            print('{}\t{}'.format(compound, compound_name(compound.name)))

        logger.info('Blocked compounds: {}'.format(count))

    def run_sink_check(self, model, solver, threshold, implicit_sinks=True):
        """Run sink production check method."""
        prob = solver.create_problem()

        # Create flux variables
        v = prob.namespace()
        for reaction_id in model.reactions:
            lower, upper = model.limits[reaction_id]
            v.define([reaction_id], lower=lower, upper=upper)

        # Build mass balance constraints
        massbalance_lhs = {compound: 0 for compound in model.compounds}
        for spec, value in iteritems(model.matrix):
            compound, reaction_id = spec
            massbalance_lhs[compound] += v(reaction_id) * value

        mass_balance_constrs = {}
        for compound, lhs in iteritems(massbalance_lhs):
            if implicit_sinks:
                # The constraint is merely >0 meaning that we have implicit
                # sinks for all compounds.
                prob.add_linear_constraints(lhs >= 0)
            else:
                # Save these constraints so we can temporarily remove them
                # to create a sink.
                c, = prob.add_linear_constraints(lhs == 0)
                mass_balance_constrs[compound] = c

        for compound, lhs in sorted(iteritems(massbalance_lhs)):
            if not implicit_sinks:
                mass_balance_constrs[compound].delete()

            prob.set_objective(lhs)
            try:
                result = prob.solve(lp.ObjectiveSense.Maximize)
            except lp.SolverError as e:
                logger.warning('Failed to solve for compound: {} ({})'.format(
                    compound, e))

            if result.get_value(lhs) < threshold:
                yield compound

            if not implicit_sinks:
                # Restore mass balance constraint.
                c, = prob.add_linear_constraints(lhs == 0)
                mass_balance_constrs[compound] = c

    def run_reaction_production_check(self, model, solver, threshold,
                                      implicit_sinks=True):
        """Run reaction production check method."""
        prob = solver.create_problem()

        # Create flux variables
        v = prob.namespace()
        for reaction_id in model.reactions:
            lower, upper = model.limits[reaction_id]
            v.define([reaction_id], lower=lower, upper=upper)

        # Build mass balance constraints
        massbalance_lhs = {compound: 0 for compound in model.compounds}
        for spec, value in iteritems(model.matrix):
            compound, reaction_id = spec
            massbalance_lhs[compound] += v(reaction_id) * value

        # Create production variables and apply constraints
        for compound, lhs in iteritems(massbalance_lhs):
            if implicit_sinks:
                # The constraint is merely >0 meaning that we have implicit
                # sinks for all compounds.
                prob.add_linear_constraints(lhs >= 0)
            else:
                prob.add_linear_constraints(lhs == 0)

        confirmed_production = set()
        for reaction in model.reactions:
            if all(c in confirmed_production for c, _ in
                   model.get_reaction_values(reaction)):
                continue

            prob.set_objective(v(reaction))
            for sense in (lp.ObjectiveSense.Maximize,
                          lp.ObjectiveSense.Minimize):
                try:
                    result = prob.solve(sense)
                except lp.SolverError as e:
                    self.fail(
                        'Failed to solve for compound, reaction: {}, {}:'
                        ' {}'.format(compound, reaction, e))

                flux = result.get_value(v(reaction))
                for compound, value in model.get_reaction_values(reaction):
                    if compound in confirmed_production:
                        continue

                    production = 0
                    if sense == lp.ObjectiveSense.Maximize and flux > 0:
                        production = float(value) * flux
                    elif sense == lp.ObjectiveSense.Minimize and flux < 0:
                        production = float(value) * flux

                    if production >= threshold:
                        confirmed_production.add(compound)

        for compound in sorted(model.compounds):
            if compound not in confirmed_production:
                yield compound

    def _log_epsilon_and_fail(self, epsilon, exc):
        msg = ('Finding blocked compounds failed with epsilon set to {}. Try'
               ' lowering the epsilon value to reduce artifical constraints'
               ' on the model.'.format(epsilon))
        self.fail(msg, exc)
